import numpy as np

from yt._maintenance.ipython_compat import IS_IPYTHON
from yt.config import ytcfg
from yt.funcs import mylog
from yt.units.yt_array import YTQuantity
from yt.utilities import png_writer as pw
from yt.utilities.exceptions import YTNotInsideNotebook
from yt.utilities.lib import image_utilities as au
from yt.visualization.color_maps import get_colormap_lut

from ._commons import get_canvas, validate_image_name


def scale_image(image, mi=None, ma=None):
    r"""Scale an image ([NxNxM] where M = 1-4) to be uint8 and values scaled
    from [0,255].

    Parameters
    ----------
    image : array_like or tuple of image info

    Examples
    --------

        >>> image = scale_image(image)

        >>> image = scale_image(image, min=0, max=1000)
    """
    if isinstance(image, np.ndarray) and image.dtype == np.uint8:
        return image
    if isinstance(image, (tuple, list)):
        image, mi, ma = image
    if mi is None:
        mi = image.min()
    if ma is None:
        ma = image.max()
    image = np.clip((image - mi) / (ma - mi) * 255, 0, 255).astype("uint8")
    return image


def multi_image_composite(
    fn, red_channel, blue_channel, green_channel=None, alpha_channel=None
):
    r"""Write an image with different color channels corresponding to different
    quantities.

    Accepts at least a red and a blue array, of shape (N,N) each, that are
    optionally scaled and composited into a final image, written into `fn`.
    Can also accept green and alpha.

    Parameters
    ----------
    fn : string
        Filename to save
    red_channel : array_like or tuple of image info
        Array, of shape (N,N), to be written into the red channel of the output
        image.  If not already uint8, will be converted (and scaled) into
        uint8.  Optionally, you can also specify a tuple that includes scaling
        information, in the form of (array_to_plot, min_value_to_scale,
        max_value_to_scale).
    blue_channel : array_like or tuple of image info
        Array, of shape (N,N), to be written into the blue channel of the output
        image.  If not already uint8, will be converted (and scaled) into
        uint8.  Optionally, you can also specify a tuple that includes scaling
        information, in the form of (array_to_plot, min_value_to_scale,
        max_value_to_scale).
    green_channel : array_like or tuple of image info, optional
        Array, of shape (N,N), to be written into the green channel of the
        output image.  If not already uint8, will be converted (and scaled)
        into uint8.  If not supplied, will be left empty.  Optionally, you can
        also specify a tuple that includes scaling information, in the form of
        (array_to_plot, min_value_to_scale, max_value_to_scale).

    alpha_channel : array_like or tuple of image info, optional
        Array, of shape (N,N), to be written into the alpha channel of the output
        image.  If not already uint8, will be converted (and scaled) into uint8.
        If not supplied, will be made fully opaque.  Optionally, you can also
        specify a tuple that includes scaling information, in the form of
        (array_to_plot, min_value_to_scale, max_value_to_scale).

    Examples
    --------

        >>> red_channel = np.log10(frb["gas", "temperature"])
        >>> blue_channel = np.log10(frb["gas", "density"])
        >>> multi_image_composite("multi_channel1.png", red_channel, blue_channel)

    """
    red_channel = scale_image(red_channel)
    blue_channel = scale_image(blue_channel)
    if green_channel is None:
        green_channel = np.zeros(red_channel.shape, dtype="uint8")
    else:
        green_channel = scale_image(green_channel)
    if alpha_channel is None:
        alpha_channel = np.zeros(red_channel.shape, dtype="uint8") + 255
    else:
        alpha_channel = scale_image(alpha_channel)
    image = np.array([red_channel, green_channel, blue_channel, alpha_channel])
    image = image.transpose().copy()  # Have to make sure it's contiguous
    pw.write_png(image, fn)


def write_bitmap(bitmap_array, filename, max_val=None, transpose=False):
    r"""Write out a bitmapped image directly to a PNG file.

    This accepts a three- or four-channel `bitmap_array`.  If the image is not
    already uint8, it will be scaled and converted.  If it is four channel,
    only the first three channels will be scaled, while the fourth channel is
    assumed to be in the range of [0,1]. If it is not four channel, a fourth
    alpha channel will be added and set to fully opaque.  The resultant image
    will be directly written to `filename` as a PNG with no colormap applied.
    `max_val` is a value used if the array is passed in as anything other than
    uint8; it will be the value used for scaling and clipping in the first
    three channels when the array is converted.  Additionally, the minimum is
    assumed to be zero; this makes it primarily suited for the results of
    volume rendered images, rather than misaligned projections.

    Parameters
    ----------
    bitmap_array : array_like
        Array of shape (N,M,3) or (N,M,4), to be written.  If it is not already
        a uint8 array, it will be scaled and converted to uint8.
    filename : string
        Filename to save to.  If None, PNG contents will be returned as a
        string.
    max_val : float, optional
        The upper limit to clip values to in the output, if converting to uint8.
        If `bitmap_array` is already uint8, this will be ignore.
    transpose : boolean, optional
        If transpose is False, we assume that the incoming bitmap_array is such
        that the first element resides in the upper-left corner.  If True, the
        first element will be placed in the lower-left corner.
    """
    if len(bitmap_array.shape) != 3 or bitmap_array.shape[-1] not in (3, 4):
        raise RuntimeError(
            "Expecting image array of shape (N,M,3) or "
            f"(N,M,4), received {str(bitmap_array.shape)}"
        )

    if bitmap_array.dtype != np.uint8:
        s1, s2 = bitmap_array.shape[:2]
        if bitmap_array.shape[-1] == 3:
            alpha_channel = np.full((s1, s2, 1), 255, dtype="uint8")
        else:
            alpha_channel = (
                (255 * bitmap_array[:, :, 3]).astype("uint8").reshape(s1, s2, 1)
            )
        if max_val is None:
            max_val = bitmap_array[:, :, :3].max()
            if max_val == 0.0:
                # avoid dividing by zero for blank images
                max_val = 1.0
        bitmap_array = np.clip(bitmap_array[:, :, :3] / max_val, 0.0, 1.0) * 255
        bitmap_array = np.concatenate(
            [bitmap_array.astype("uint8"), alpha_channel], axis=-1
        )
    if transpose:
        bitmap_array = bitmap_array.swapaxes(0, 1).copy(order="C")
    if filename is not None:
        pw.write_png(bitmap_array, filename)
    else:
        return pw.write_png_to_string(bitmap_array.copy())
    return bitmap_array


def write_image(image, filename, color_bounds=None, cmap_name=None, func=lambda x: x):
    r"""Write out a floating point array directly to a PNG file, scaling it and
    applying a colormap.

    This function will scale an image and directly call libpng to write out a
    colormapped version of that image.  It is designed for rapid-fire saving of
    image buffers generated using `yt.visualization.api.FixedResolutionBuffers`
    and the likes.

    Parameters
    ----------
    image : array_like
        This is an (unscaled) array of floating point values, shape (N,N,) to
        save in a PNG file.
    filename : string
        Filename to save as.
    color_bounds : tuple of floats, optional
        The min and max to scale between.  Outlying values will be clipped.
    cmap_name : string, optional
        An acceptable colormap.  See either yt.visualization.color_maps or
        https://scipy-cookbook.readthedocs.io/items/Matplotlib_Show_colormaps.html .
    func : function, optional
        A function to transform the buffer before applying a colormap.

    Returns
    -------
    scaled_image : uint8 image that has been saved

    Examples
    --------

    >>> sl = ds.slice(0, 0.5, "Density")
    >>> frb1 = FixedResolutionBuffer(sl, (0.2, 0.3, 0.4, 0.5), (1024, 1024))
    >>> write_image(frb1["gas", "density"], "saved.png")
    """
    if cmap_name is None:
        cmap_name = ytcfg.get("yt", "default_colormap")
    if len(image.shape) == 3:
        mylog.info("Using only channel 1 of supplied image")
        image = image[:, :, 0]
    to_plot = apply_colormap(image, color_bounds=color_bounds, cmap_name=cmap_name)
    pw.write_png(to_plot, filename)
    return to_plot


def apply_colormap(image, color_bounds=None, cmap_name=None, func=lambda x: x):
    r"""Apply a colormap to a floating point image, scaling to uint8.

    This function will scale an image and directly call libpng to write out a
    colormapped version of that image.  It is designed for rapid-fire saving of
    image buffers generated using `yt.visualization.api.FixedResolutionBuffers`
    and the likes.

    Parameters
    ----------
    image : array_like
        This is an (unscaled) array of floating point values, shape (N,N,) to
        save in a PNG file.
    color_bounds : tuple of floats, optional
        The min and max to scale between.  Outlying values will be clipped.
    cmap_name : string, optional
        An acceptable colormap.  See either yt.visualization.color_maps or
        https://scipy-cookbook.readthedocs.io/items/Matplotlib_Show_colormaps.html .
    func : function, optional
        A function to transform the buffer before applying a colormap.

    Returns
    -------
    to_plot : uint8 image with colorbar applied.

    """
    if cmap_name is None:
        cmap_name = ytcfg.get("yt", "default_colormap")
    from yt.data_objects.image_array import ImageArray

    image = ImageArray(func(image))
    if color_bounds is None:
        mi = np.nanmin(image[~np.isinf(image)]) * image.uq
        ma = np.nanmax(image[~np.isinf(image)]) * image.uq
        color_bounds = mi, ma
    else:
        color_bounds = [YTQuantity(func(c), image.units) for c in color_bounds]
    image = (image - color_bounds[0]) / (color_bounds[1] - color_bounds[0])
    to_plot = map_to_colors(image, cmap_name)
    to_plot = np.clip(to_plot, 0, 255)
    return to_plot


def map_to_colors(buff, cmap_name):
    lut = get_colormap_lut(cmap_name)

    if isinstance(cmap_name, tuple):
        # If we are using the colorbrewer maps, don't interpolate
        shape = buff.shape
        # We add float_eps so that digitize doesn't go out of bounds
        x = np.mgrid[0.0 : 1.0 + np.finfo(np.float32).eps : lut[0].shape[0] * 1j]
        inds = np.digitize(buff.ravel(), x).reshape(shape[0], shape[1])
        mapped = np.dstack([(v[inds] * 255).astype("uint8") for v in lut])
        del inds
    else:
        x = np.mgrid[0.0 : 1.0 : lut[0].shape[0] * 1j]
        mapped = np.dstack([(np.interp(buff, x, v) * 255).astype("uint8") for v in lut])
    return mapped.copy("C")


def splat_points(image, points_x, points_y, contribution=None, transposed=False):
    if contribution is None:
        contribution = 100.0
    val = contribution * 1.0 / points_x.size
    if transposed:
        points_y = 1.0 - points_y
        points_x = 1.0 - points_x
    im = image.copy()
    au.add_points_to_image(im, points_x, points_y, val)
    return im


def write_projection(
    data,
    filename,
    colorbar=True,
    colorbar_label=None,
    title=None,
    vmin=None,
    vmax=None,
    take_log=True,
    figsize=(8, 6),
    dpi=100,
    cmap_name=None,
    extent=None,
    xlabel=None,
    ylabel=None,
):
    r"""Write a projection or volume rendering to disk with a variety of
    pretty parameters such as limits, title, colorbar, etc.  write_projection
    uses the standard matplotlib interface to create the figure.  N.B. This code
    only works *after* you have created the projection using the standard
    framework (i.e. the Camera interface or off_axis_projection).

    Accepts an NxM sized array representing the projection itself as well
    as the filename to which you will save this figure.  Note that the final
    resolution of your image will be a product of dpi/100 * figsize.

    Parameters
    ----------
    data : array_like
        image array as output by off_axis_projection or camera.snapshot()
    filename : string
        the filename where the data will be saved
    colorbar : boolean
        do you want a colorbar generated to the right of the image?
    colorbar_label : string
        the label associated with your colorbar
    title : string
        the label at the top of the figure
    vmin : float or None
        the lower limit of the zaxis (part of matplotlib api)
    vmax : float or None
        the lower limit of the zaxis (part of matplotlib api)
    take_log : boolean
        plot the log of the data array (and take the log of the limits if set)?
    figsize : array_like
        width, height in inches of final image
    dpi : int
        final image resolution in pixels / inch
    cmap_name : string
        The name of the colormap.

    Examples
    --------

    >>> image = off_axis_projection(ds, c, L, W, N, "Density", no_ghost=False)
    >>> write_projection(
    ...     image,
    ...     "test.png",
    ...     colorbar_label="Column Density (cm$^{-2}$)",
    ...     title="Offaxis Projection",
    ...     vmin=1e-5,
    ...     vmax=1e-3,
    ...     take_log=True,
    ... )
    """
    if cmap_name is None:
        cmap_name = ytcfg.get("yt", "default_colormap")
    import matplotlib.colors
    import matplotlib.figure

    # If this is rendered as log, then apply now.
    if take_log:
        norm_cls = matplotlib.colors.LogNorm
    else:
        norm_cls = matplotlib.colors.Normalize
    norm = norm_cls(vmin=vmin, vmax=vmax)

    # Create the figure and paint the data on
    fig = matplotlib.figure.Figure(figsize=figsize)
    ax = fig.add_subplot(111)

    cax = ax.imshow(
        data.to_ndarray(),
        norm=norm,
        extent=extent,
        cmap=cmap_name,
    )

    if title:
        ax.set_title(title)

    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)

    # Suppress the x and y pixel counts
    if extent is None:
        ax.set_xticks(())
        ax.set_yticks(())

    # Add a color bar and label if requested
    if colorbar:
        cbar = fig.colorbar(cax)
        if colorbar_label:
            cbar.ax.set_ylabel(colorbar_label)

    filename = validate_image_name(filename)
    canvas = get_canvas(fig, filename)

    mylog.info("Saving plot %s", filename)
    fig.tight_layout()

    canvas.print_figure(filename, dpi=dpi)
    return filename


def display_in_notebook(image, max_val=None):
    """
    A helper function to display images in an IPython notebook

    Must be run from within an IPython notebook, or else it will raise
    a YTNotInsideNotebook exception.

    Parameters
    ----------
    image : array_like
        This is an (unscaled) array of floating point values, shape (N,N,3) or
        (N,N,4) to display in the notebook. The first three channels will be
        scaled automatically.
    max_val : float, optional
        The upper limit to clip values of the image.  Only applies to the first
        three channels.
    """

    if IS_IPYTHON:
        from IPython.core.displaypub import publish_display_data

        data = write_bitmap(image, None, max_val=max_val)
        publish_display_data(
            data={"image/png": data},
            source="yt.visualization.image_writer.display_in_notebook",
        )
    else:
        raise YTNotInsideNotebook



# distutils: language = c++
"""
Cython tools for working with the PyKDTree particle KDTree.



"""


import numpy as np

cimport cython
cimport numpy as np
from cpython.exc cimport PyErr_CheckSignals
from libc.math cimport sqrt

from yt.utilities.lib.cykdtree.kdtree cimport KDTree, Node, PyKDTree, uint32_t, uint64_t

from yt.funcs import get_pbar

from yt.geometry.particle_deposit cimport get_kernel_func, kernel_func
from yt.utilities.lib.bounded_priority_queue cimport BoundedPriorityQueue, NeighborList


cdef int CHUNKSIZE = 4096

# This structure allows the nearest neighbor finding to consider a subset of
# spatial dimensions, i.e the spatial separation in the x and z coordinates
# could be consider by using set_axes_range(axes, 1), this would cause the while
# loops to skip the y dimensions, without the performance hit of an if statement
cdef struct axes_range:
    int start
    int stop
    int step

# skipaxis: x=0, y=1, z=2
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef int set_axes_range(axes_range *axes, int skipaxis):
    axes.start = 0
    axes.stop = 3
    axes.step = 1
    if skipaxis == 0:
        axes.start = 1
    if skipaxis == 1:
        axes.step = 2
    if skipaxis == 2:
        axes.stop = 2
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def generate_smoothing_length(np.float64_t[:, ::1] tree_positions,
                              PyKDTree kdtree, int n_neighbors):
    """Calculate array of distances to the nth nearest neighbor

    Parameters
    ----------

    tree_positions: arrays of floats with shape (n_particles, 3)
        The positions of particles in kdtree sorted order. Currently assumed
        to be 3D positions.
    kdtree: A PyKDTree instance
        A kdtree to do nearest neighbors searches with
    n_neighbors: The neighbor number to calculate the distance to

    Returns
    -------

    smoothing_lengths: arrays of floats with shape (n_particles, )
        The calculated smoothing lengths

    """
    cdef int i
    cdef KDTree * c_tree = kdtree._tree
    cdef int n_particles = tree_positions.shape[0]
    cdef np.float64_t * pos
    cdef np.float64_t[:] smoothing_length = np.empty(n_particles)
    cdef BoundedPriorityQueue queue = BoundedPriorityQueue(n_neighbors)

    # We are using all spatial dimensions
    cdef axes_range axes
    set_axes_range(&axes, -1)

    pbar = get_pbar("Generate smoothing length", n_particles)
    with nogil:
        for i in range(n_particles):
            # Reset queue to "empty" state, doing it this way avoids
            # needing to reallocate memory
            queue.size = 0

            if i % CHUNKSIZE == 0:
                with gil:
                    pbar.update(i-1)
                    PyErr_CheckSignals()

            pos = &(tree_positions[i, 0])
            find_neighbors(pos, tree_positions, queue, c_tree, i, &axes)

            smoothing_length[i] = sqrt(queue.heap_ptr[0])

    pbar.update(n_particles-1)
    pbar.finish()
    return np.asarray(smoothing_length)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def estimate_density(np.float64_t[:, ::1] tree_positions, np.float64_t[:] mass,
                      np.float64_t[:] smoothing_length,
                      PyKDTree kdtree, kernel_name="cubic"):
    """Estimate density using SPH gather method.

    Parameters
    ----------

    tree_positions: array of floats with shape (n_particles, 3)
        The positions of particles in kdtree sorted order. Currently assumed
        to be 3D positions.
    mass: array of floats with shape (n_particles)
        The masses of particles in kdtree sorted order.
    smoothing_length: array of floats with shape (n_particles)
        The smoothing lengths of particles in kdtree sorted order.
    kdtree: A PyKDTree instance
        A kdtree to do nearest neighbors searches with.
    kernel_name: str
        The name of the kernel function to use in density estimation.

    Returns
    -------

    density: array of floats with shape (n_particles)
        The calculated density.

    """
    cdef int i, j, k
    cdef KDTree * c_tree = kdtree._tree
    cdef int n_particles = tree_positions.shape[0]
    cdef np.float64_t h_i2, ih_i2, q_ij
    cdef np.float64_t * pos
    cdef np.float64_t[:] density = np.empty(n_particles)
    cdef kernel_func kernel = get_kernel_func(kernel_name)
    cdef NeighborList nblist = NeighborList()

    # We are using all spatial dimensions
    cdef axes_range axes
    set_axes_range(&axes, -1)

    pbar = get_pbar("Estimating density", n_particles)
    with nogil:
        for i in range(n_particles):
            # Reset list to "empty" state, doing it this way avoids
            # needing to reallocate memory
            nblist.size = 0

            if i % CHUNKSIZE == 0:
                with gil:
                    pbar.update(i - 1)
                    PyErr_CheckSignals()

            pos = &(tree_positions[i, 0])
            h_i2 = smoothing_length[i] ** 2
            find_neighbors_ball(pos, h_i2, tree_positions, nblist, c_tree, i, &axes)
            ih_i2 = 1.0 / h_i2

            # See eq. 10 of Price 2012
            density[i] = mass[i] * kernel(0)
            for k in range(nblist.size):
                j = nblist.pids[k]
                q_ij = sqrt(nblist.data[k] * ih_i2)
                density[i] += mass[j] * kernel(q_ij)

    pbar.update(n_particles - 1)
    pbar.finish()
    return np.asarray(density)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int find_neighbors(np.float64_t * pos, np.float64_t[:, ::1] tree_positions,
                        BoundedPriorityQueue queue, KDTree * c_tree,
                        uint64_t skipidx, axes_range * axes) except -1 nogil:
    cdef Node* leafnode

    # Make an initial guess based on the closest node
    leafnode = c_tree.search(&pos[0])
    process_node_points(leafnode, queue, tree_positions, pos, skipidx, axes)

    # Traverse the rest of the kdtree to finish the neighbor list
    find_knn(c_tree.root, queue, tree_positions, pos, leafnode.leafid, skipidx,
             axes)

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int find_knn(Node* node,
                  BoundedPriorityQueue queue,
                  np.float64_t[:, ::1] tree_positions,
                  np.float64_t* pos,
                  uint32_t skipleaf,
                  uint64_t skipidx,
                  axes_range * axes,
                  ) except -1 nogil:
    # if we aren't a leaf then we keep traversing until we find a leaf, else we
    # we actually begin to check the leaf
    if not node.is_leaf:
        if not cull_node(node.less, pos, queue, skipleaf, axes):
            find_knn(node.less, queue, tree_positions, pos, skipleaf, skipidx,
                     axes)
        if not cull_node(node.greater, pos, queue, skipleaf, axes):
            find_knn(node.greater, queue, tree_positions, pos, skipleaf,
                     skipidx, axes)
    else:
        if not cull_node(node, pos, queue, skipleaf, axes):
            process_node_points(node, queue, tree_positions, pos, skipidx,
                                axes)
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int cull_node(Node* node,
                          np.float64_t* pos,
                          BoundedPriorityQueue queue,
                          uint32_t skipleaf,
                          axes_range * axes,
                          ) except -1 nogil:
    cdef int k
    cdef np.float64_t v
    cdef np.float64_t tpos, ndist = 0

    if node.leafid == skipleaf:
        return True

    k = axes.start
    while k < axes.stop:
        v = pos[k]
        if v < node.left_edge[k]:
            tpos = node.left_edge[k] - v
        elif v > node.right_edge[k]:
            tpos = v - node.right_edge[k]
        else:
            tpos = 0
        ndist += tpos*tpos
        k += axes.step

    return (ndist > queue.heap[0] and queue.size == queue.max_elements)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int process_node_points(Node* node,
                                    BoundedPriorityQueue queue,
                                    np.float64_t[:, ::1] positions,
                                    np.float64_t* pos,
                                    int skipidx,
                                    axes_range * axes,
                                    ) except -1 nogil:
    cdef uint64_t i, k
    cdef np.float64_t tpos, sq_dist
    for i in range(node.left_idx, node.left_idx + node.children):
        if i == skipidx:
            continue

        sq_dist = 0.0

        k = axes.start
        while k < axes.stop:
            tpos = positions[i, k] - pos[k]
            sq_dist += tpos*tpos
            k += axes.step

        queue.add_pid(sq_dist, i)

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int find_neighbors_ball(np.float64_t * pos, np.float64_t r2,
                             np.float64_t[:, ::1] tree_positions,
                             NeighborList nblist, KDTree * c_tree,
                             uint64_t skipidx, axes_range * axes
                             ) except -1 nogil:
    """Find neighbors within a ball."""
    cdef Node* leafnode

    # Make an initial guess based on the closest node
    leafnode = c_tree.search(&pos[0])
    process_node_points_ball(leafnode, nblist, tree_positions, pos, r2, skipidx, axes)

    # Traverse the rest of the kdtree to finish the neighbor list
    find_ball(c_tree.root, nblist, tree_positions, pos, r2, leafnode.leafid,
              skipidx, axes)

    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int find_ball(Node* node,
                   NeighborList nblist,
                   np.float64_t[:, ::1] tree_positions,
                   np.float64_t* pos,
                   np.float64_t r2,
                   uint32_t skipleaf,
                   uint64_t skipidx,
                   axes_range * axes,
                   ) except -1 nogil:
    """Traverse the k-d tree to process leaf nodes."""
    if not node.is_leaf:
        if not cull_node_ball(node.less, pos, r2, skipleaf, axes):
            find_ball(node.less, nblist, tree_positions, pos, r2, skipleaf,
                      skipidx, axes)
        if not cull_node_ball(node.greater, pos, r2, skipleaf, axes):
            find_ball(node.greater, nblist, tree_positions, pos, r2, skipleaf,
                      skipidx, axes)
    else:
        if not cull_node_ball(node, pos, r2, skipleaf, axes):
            process_node_points_ball(node, nblist, tree_positions, pos, r2,
                                     skipidx, axes)
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int cull_node_ball(Node* node,
                               np.float64_t* pos,
                               np.float64_t r2,
                               uint32_t skipleaf,
                               axes_range * axes,
                               ) except -1 nogil:
    """Check if the node does not intersect with the ball at all."""
    cdef int k
    cdef np.float64_t v
    cdef np.float64_t tpos, ndist = 0

    if node.leafid == skipleaf:
        return True

    k = axes.start
    while k < axes.stop:
        v = pos[k]
        if v < node.left_edge[k]:
            tpos = node.left_edge[k] - v
        elif v > node.right_edge[k]:
            tpos = v - node.right_edge[k]
        else:
            tpos = 0
        ndist += tpos*tpos
        k += axes.step

    return ndist > r2

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int process_node_points_ball(Node* node,
                                         NeighborList nblist,
                                         np.float64_t[:, ::1] positions,
                                         np.float64_t* pos,
                                         np.float64_t r2,
                                         int skipidx,
                                         axes_range * axes,
                                         ) except -1 nogil:
    """Add points from the leaf node within the ball to the neighbor list."""
    cdef uint64_t i, k
    cdef np.float64_t tpos, sq_dist
    for i in range(node.left_idx, node.left_idx + node.children):
        if i == skipidx:
            continue

        sq_dist = 0.0

        k = axes.start
        while k < axes.stop:
            tpos = positions[i, k] - pos[k]
            sq_dist += tpos*tpos
            k += axes.step

        if (sq_dist < r2):
            nblist.add_pid(sq_dist, i)

    return 0



import os

import numpy as np

from yt.utilities.io_handler import BaseParticleIOHandler

from .definitions import halo_dts


class IOHandlerRockstarBinary(BaseParticleIOHandler):
    _dataset_type = "rockstar_binary"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._halo_dt = halo_dts[self.ds.parameters["format_revision"]]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.

        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        ptype = "halos"
        for data_file in self._sorted_chunk_iterator(chunks):
            pcount = data_file.header["num_halos"]
            if pcount == 0:
                continue
            with open(data_file.filename, "rb") as f:
                pos = data_file._get_particle_positions(ptype, f=f)
                yield "halos", (pos[:, i] for i in range(3)), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        for data_file in self._sorted_chunk_iterator(chunks):
            si, ei = data_file.start, data_file.end
            pcount = data_file.header["num_halos"]
            if pcount == 0:
                continue
            with open(data_file.filename, "rb") as f:
                for ptype, field_list in sorted(ptf.items()):
                    pos = data_file._get_particle_positions(ptype, f=f)
                    x, y, z = (pos[:, i] for i in range(3))
                    mask = selector.select_points(x, y, z, 0.0)
                    del x, y, z
                    f.seek(data_file._position_offset, os.SEEK_SET)
                    halos = np.fromfile(f, dtype=self._halo_dt, count=pcount)
                    if mask is None:
                        continue
                    for field in field_list:
                        data = halos[field][si:ei][mask].astype("float64")
                        yield (ptype, field), data

    def _yield_coordinates(self, data_file):
        # Just does halos
        pcount = data_file.header["num_halos"]
        with open(data_file.filename, "rb") as f:
            f.seek(data_file._position_offset, os.SEEK_SET)
            halos = np.fromfile(f, dtype=self._halo_dt, count=pcount)
            pos = np.empty((halos.size, 3), dtype="float64")
            pos[:, 0] = halos["particle_position_x"]
            pos[:, 1] = halos["particle_position_y"]
            pos[:, 2] = halos["particle_position_z"]
            yield "halos", pos

    def _count_particles(self, data_file):
        nhalos = data_file.header["num_halos"]
        si, ei = data_file.start, data_file.end
        if None not in (si, ei):
            nhalos = np.clip(nhalos - si, 0, ei - si)
        return {"halos": nhalos}

    def _identify_fields(self, data_file):
        fields = [("halos", f) for f in self._halo_dt.fields if "padding" not in f]
        return fields, {}



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

import yt.units as u
from yt.testing import fake_random_ds
from yt.visualization.volume_rendering.api import Scene

valid_lens_types = [
    "plane-parallel",
    "perspective",
    "stereo-perspective",
    "fisheye",
    "spherical",
    "stereo-spherical",
]


def test_scene_and_camera_attributes():
    ds = fake_random_ds(64, length_unit=2, bbox=np.array([[-1, 1], [-1, 1], [-1, 1]]))
    sc = Scene()
    cam = sc.add_camera(ds)

    # test that initial values are correct in code units
    assert_equal(cam.width, ds.arr([3, 3, 3], "code_length"))
    assert_equal(cam.position, ds.arr([1, 1, 1], "code_length"))
    assert_equal(cam.focus, ds.arr([0, 0, 0], "code_length"))

    # test setting the attributes in various ways

    attribute_values = [
        (
            1,
            ds.arr([2, 2, 2], "code_length"),
        ),
        (
            [1],
            ds.arr([2, 2, 2], "code_length"),
        ),
        (
            [1, 2],
            RuntimeError,
        ),
        (
            [1, 1, 1],
            ds.arr([2, 2, 2], "code_length"),
        ),
        (
            (1, "code_length"),
            ds.arr([1, 1, 1], "code_length"),
        ),
        (
            ((1, "code_length"), (1, "code_length")),
            RuntimeError,
        ),
        (
            ((1, "cm"), (2, "cm"), (3, "cm")),
            ds.arr([0.5, 1, 1.5], "code_length"),
        ),
        (
            2 * u.cm,
            ds.arr([1, 1, 1], "code_length"),
        ),
        (
            ds.arr(2, "cm"),
            ds.arr([1, 1, 1], "code_length"),
        ),
        (
            [2 * u.cm],
            ds.arr([1, 1, 1], "code_length"),
        ),
        (
            [1, 2, 3] * u.cm,
            ds.arr([0.5, 1, 1.5], "code_length"),
        ),
        (
            [1, 2] * u.cm,
            RuntimeError,
        ),
        (
            [u.cm * w for w in [1, 2, 3]],
            ds.arr([0.5, 1, 1.5], "code_length"),
        ),
    ]

    # define default values to avoid accidentally setting focus = position
    default_values = {
        "focus": [0, 0, 0],
        "position": [4, 4, 4],
        "width": [1, 1, 1],
    }
    attribute_list = list(default_values.keys())

    for attribute in attribute_list:
        for other_attribute in [a for a in attribute_list if a != attribute]:
            setattr(cam, other_attribute, default_values[other_attribute])
        for attribute_value, expected_result in attribute_values:
            try:
                # test properties
                setattr(cam, attribute, attribute_value)
                assert_almost_equal(getattr(cam, attribute), expected_result)
            except RuntimeError:
                assert expected_result is RuntimeError

            try:
                # test setters/getters
                getattr(cam, f"set_{attribute}")(attribute_value)
                assert_almost_equal(getattr(cam, f"get_{attribute}")(), expected_result)
            except RuntimeError:
                assert expected_result is RuntimeError

    resolution_values = (
        (
            512,
            (512, 512),
        ),
        (
            (512, 512),
            (512, 512),
        ),
        (
            (256, 512),
            (256, 512),
        ),
        ((256, 256, 256), RuntimeError),
    )

    for resolution_value, expected_result in resolution_values:
        try:
            # test properties
            cam.resolution = resolution_value
            assert_equal(cam.resolution, expected_result)
        except RuntimeError:
            assert expected_result is RuntimeError

        try:
            # test setters/getters
            cam.set_resolution(resolution_value)
            assert_equal(cam.get_resolution(), expected_result)
        except RuntimeError:
            assert expected_result is RuntimeError

    for lens_type in valid_lens_types:
        cam.set_lens(lens_type)

    # See issue #1287
    cam.focus = [0, 0, 0]
    cam_pos = [1, 0, 0]
    north_vector = [0, 1, 0]
    cam.set_position(cam_pos, north_vector)
    cam_pos = [0, 1, 0]
    north_vector = [0, 0, 1]
    cam.set_position(cam_pos, north_vector)



import os.path

from numpy.testing import assert_equal

from yt.frontends.rockstar.api import RockstarDataset
from yt.testing import ParticleSelectionComparison, requires_file
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    data_dir_load,
    requires_ds,
)

_fields = (
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
    ("all", "particle_mass"),
)

r1 = "rockstar_halos/halos_0.0.bin"


@requires_ds(r1)
def test_fields_r1():
    ds = data_dir_load(r1)
    assert_equal(str(ds), os.path.basename(r1))
    for field in _fields:
        yield FieldValuesTest(r1, field, particle_type=True)


@requires_file(r1)
def test_RockstarDataset():
    assert isinstance(data_dir_load(r1), RockstarDataset)


@requires_file(r1)
def test_particle_selection():
    ds = data_dir_load(r1)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


@requires_file(r1)
def test_halo_loading():
    ds = data_dir_load(r1)

    for halo_id, Npart in zip(
        ds.r["halos", "particle_identifier"],
        ds.r["halos", "num_p"],
        strict=False,
    ):
        halo = ds.halo("halos", halo_id)
        assert halo is not None

        # Try accessing properties
        halo.position
        halo.velocity
        halo.mass

        # Make sure we can access the member particles
        assert_equal(len(halo.member_ids), Npart)



import matplotlib.pyplot as plt
import pytest

from yt.config import ytcfg
from yt.testing import (
    fake_hexahedral_ds,
    fake_tetrahedral_ds,
    requires_file,
    requires_module,
)
from yt.utilities.answer_testing.framework import data_dir_load, data_dir_load_v2
from yt.visualization.volume_rendering.api import MeshSource, Scene, create_scene
from yt.visualization.volume_rendering.render_source import set_raytracing_engine


@pytest.fixture
@requires_module("pyembree")
def with_pyembree_ray_tracing_engine():
    old = ytcfg["yt", "ray_tracing_engine"]

    # the @requires_module decorator only guards against pyembree not being
    # available for import, but it might not be installed properly regardless
    # so we need to be extra careful not to run tests with the default engine
    try:
        set_raytracing_engine("embree")
    except UserWarning as exc:
        pytest.skip(str(exc))
    else:
        if ytcfg["yt", "ray_tracing_engine"] != "embree":
            pytest.skip("Error while setting embree raytracing engine")

    yield

    set_raytracing_engine(old)


@pytest.fixture
def with_default_ray_tracing_engine():
    old = ytcfg["yt", "ray_tracing_engine"]
    set_raytracing_engine("yt")
    yield
    set_raytracing_engine(old)


@pytest.mark.usefixtures("with_default_ray_tracing_engine")
class TestMeshRenderDefaultEngine:
    @classmethod
    def setup_class(cls):
        cls.images = {}

        cls.ds_t = fake_tetrahedral_ds()
        for field in cls.ds_t.field_list:
            if field[0] == "all":
                continue
            sc = Scene()
            sc.add_source(MeshSource(cls.ds_t, field))
            sc.add_camera()
            im = sc.render()
            cls.images["tetrahedral_" + "_".join(field)] = im

        cls.ds_h = fake_hexahedral_ds()
        for field in cls.ds_t.field_list:
            if field[0] == "all":
                continue
            sc = Scene()
            sc.add_source(MeshSource(cls.ds_h, field))
            sc.add_camera()
            im = sc.render()
            cls.images["hexahedral_" + "_".join(field)] = im

    @pytest.mark.parametrize("kind", ["tetrahedral", "hexahedral"])
    @pytest.mark.parametrize("fname", ["elem", "test"])
    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine(self, kind, fname):
        fig, ax = plt.subplots()
        ax.imshow(self.images[f"{kind}_connect1_{fname}"])
        return fig


@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
class TestMeshRenderPyembreeEngine:
    @classmethod
    def setup_class(cls):
        cls.images = {}

        cls.ds_t = fake_tetrahedral_ds()
        for field in cls.ds_t.field_list:
            if field[0] == "all":
                continue
            sc = Scene()
            sc.add_source(MeshSource(cls.ds_t, field))
            sc.add_camera()
            im = sc.render()
            cls.images["tetrahedral_" + "_".join(field)] = im

        cls.ds_h = fake_hexahedral_ds()
        for field in cls.ds_t.field_list:
            if field[0] == "all":
                continue
            sc = Scene()
            sc.add_source(MeshSource(cls.ds_h, field))
            sc.add_camera()
            im = sc.render()
            cls.images["hexahedral_" + "_".join(field)] = im

    @pytest.mark.parametrize("kind", ["tetrahedral", "hexahedral"])
    @pytest.mark.parametrize("fname", ["elem", "test"])
    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine(self, kind, fname):
        fig, ax = plt.subplots()
        ax.imshow(self.images[f"{kind}_connect1_{fname}"])
        return fig


hex8 = "MOOSE_sample_data/out.e-s010"


@pytest.mark.usefixtures("with_default_ray_tracing_engine")
class TestHex8DefaultEngine:
    @requires_file(hex8)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(hex8, step=-1)
        cls.images = {}
        for field in [("connect1", "diffused"), ("connect2", "convected")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine_hex8_connect1_diffused(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_diffused"])
        return fig

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine_hex8_connect2_convected(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect2_convected"])
        return fig


@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
class TestHex8PyembreeEngine:
    @requires_file(hex8)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(hex8, step=-1)
        cls.images = {}
        for field in [("connect1", "diffused"), ("connect2", "convected")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine_hex8_connect1_diffused(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_diffused"])
        return fig

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine_hex8_connect2_convected(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect2_convected"])
        return fig


tet4 = "MOOSE_sample_data/high_order_elems_tet4_refine_out.e"


@pytest.mark.usefixtures("with_default_ray_tracing_engine")
class TestTet4DefaultEngine:
    @requires_file(tet4)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(tet4, step=-1)
        cls.images = {}
        for field in [("connect1", "u")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine_tet4(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_u"])
        return fig


@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
class TestTet4PyembreeEngine:
    @requires_file(tet4)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(tet4, step=-1)
        cls.images = {}
        for field in [("connect1", "u")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine_tet4(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_u"])
        return fig


hex20 = "MOOSE_sample_data/mps_out.e"


@pytest.mark.usefixtures("with_default_ray_tracing_engine")
class TestHex20DefaultEngine:
    @requires_file(hex20)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(hex20, step=-1)
        cls.images = {}
        for field in [("connect2", "temp")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine_hex20(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect2_temp"])
        return fig


@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
class TestHex20PyembreeEngine:
    @requires_file(hex20)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(hex20, step=-1)
        cls.images = {}
        for field in [("connect2", "temp")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine_hex20(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect2_temp"])
        return fig


wedge6 = "MOOSE_sample_data/wedge_out.e"


@pytest.mark.usefixtures("with_default_ray_tracing_engine")
class TestWedge6DefaultEngine:
    @requires_file(wedge6)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(wedge6, step=-1)
        cls.images = {}
        for field in [("connect1", "diffused")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine_wedge6(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_diffused"])
        return fig


@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
class TestWedge6PyembreeEngine:
    @requires_file(wedge6)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(wedge6, step=-1)
        cls.images = {}
        for field in [("connect1", "diffused")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine_wedge6(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_diffused"])
        return fig


tet10 = "SecondOrderTets/tet10_unstructured_out.e"


@pytest.mark.usefixtures("with_default_ray_tracing_engine")
class TestTet10DefaultEngine:
    @requires_file(tet10)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(tet10, step=-1)
        cls.images = {}
        for field in [("connect1", "uz")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_default_engine_tet10(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_uz"])
        return fig


@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
class TestTet10PyembreeEngine:
    @requires_file(tet10)
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2(tet10, step=-1)
        cls.images = {}
        for field in [("connect1", "uz")]:
            sc = create_scene(cls.ds, field)
            im = sc.render()
            cls.images["_".join(field)] = im

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_mesh_render_pyembree_engine_tet10(self):
        fig, ax = plt.subplots()
        ax.imshow(self.images["connect1_uz"])
        return fig


@requires_file(hex8)
@pytest.mark.usefixtures("with_default_ray_tracing_engine")
@pytest.mark.mpl_image_compare
def test_perspective_mesh_render_default():
    ds = data_dir_load(hex8)
    sc = create_scene(ds, ("connect2", "diffused"))
    cam = sc.add_camera(ds, lens_type="perspective")
    cam.focus = ds.arr([0.0, 0.0, 0.0], "code_length")
    cam_pos = ds.arr([-4.5, 4.5, -4.5], "code_length")
    north_vector = ds.arr([0.0, -1.0, -1.0], "dimensionless")
    cam.set_position(cam_pos, north_vector)
    cam.resolution = (800, 800)
    im = sc.render()

    fig, ax = plt.subplots()
    ax.imshow(im)
    return fig


@requires_file(hex8)
@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
@pytest.mark.mpl_image_compare
def test_perspective_mesh_render_pyembree():
    ds = data_dir_load(hex8)
    sc = create_scene(ds, ("connect2", "diffused"))
    cam = sc.add_camera(ds, lens_type="perspective")
    cam.focus = ds.arr([0.0, 0.0, 0.0], "code_length")
    cam_pos = ds.arr([-4.5, 4.5, -4.5], "code_length")
    north_vector = ds.arr([0.0, -1.0, -1.0], "dimensionless")
    cam.set_position(cam_pos, north_vector)
    cam.resolution = (800, 800)
    im = sc.render()

    fig, ax = plt.subplots()
    ax.imshow(im)
    return fig


@requires_file(hex8)
@pytest.mark.usefixtures("with_default_ray_tracing_engine")
@pytest.mark.mpl_image_compare
def test_composite_mesh_render_default():
    ds = data_dir_load(hex8)
    sc = Scene()
    cam = sc.add_camera(ds)
    cam.focus = ds.arr([0.0, 0.0, 0.0], "code_length")
    cam.set_position(
        ds.arr([-3.0, 3.0, -3.0], "code_length"),
        ds.arr([0.0, -1.0, 0.0], "dimensionless"),
    )
    cam.set_width = ds.arr([8.0, 8.0, 8.0], "code_length")
    cam.resolution = (800, 800)
    ms1 = MeshSource(ds, ("connect1", "diffused"))
    ms2 = MeshSource(ds, ("connect2", "diffused"))
    sc.add_source(ms1)
    sc.add_source(ms2)
    im = sc.render()

    fig, ax = plt.subplots()
    ax.imshow(im)
    return fig


@requires_file(hex8)
@pytest.mark.usefixtures("with_pyembree_ray_tracing_engine")
@pytest.mark.mpl_image_compare
def test_composite_mesh_render_pyembree():
    ds = data_dir_load(hex8)
    sc = Scene()
    cam = sc.add_camera(ds)
    cam.focus = ds.arr([0.0, 0.0, 0.0], "code_length")
    cam.set_position(
        ds.arr([-3.0, 3.0, -3.0], "code_length"),
        ds.arr([0.0, -1.0, 0.0], "dimensionless"),
    )
    cam.set_width = ds.arr([8.0, 8.0, 8.0], "code_length")
    cam.resolution = (800, 800)
    ms1 = MeshSource(ds, ("connect1", "diffused"))
    ms2 = MeshSource(ds, ("connect2", "diffused"))
    sc.add_source(ms1)
    sc.add_source(ms2)
    im = sc.render()

    fig, ax = plt.subplots()
    ax.imshow(im)
    return fig



import numpy as np

from yt.testing import fake_random_ds
from yt.utilities.flagging_methods import flagging_method_registry


def test_over_density():
    ds = fake_random_ds(64)
    ds.index
    od_flag = flagging_method_registry["overdensity"](0.75)
    criterion = ds.index.grids[0]["gas", "density"] > 0.75
    assert np.all(od_flag(ds.index.grids[0]) == criterion)



import numpy as np
from numpy.testing import assert_almost_equal

from yt.loaders import load_uniform_grid
from yt.utilities.physical_constants import mu_0


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_magnetic_fields():
    ddims = (16, 16, 16)
    data1 = {
        "magnetic_field_x": (np.random.random(size=ddims), "T"),
        "magnetic_field_y": (np.random.random(size=ddims), "T"),
        "magnetic_field_z": (np.random.random(size=ddims), "T"),
    }
    data2 = {}
    for field in data1:
        data2[field] = (data1[field][0] * 1.0e4, "gauss")

    ds0 = load_uniform_grid(data1, ddims, unit_system="cgs")
    ds1 = load_uniform_grid(data1, ddims, magnetic_unit=(1.0, "T"), unit_system="cgs")
    ds2 = load_uniform_grid(data2, ddims, unit_system="mks")
    # For this test dataset, code units are cgs units
    ds3 = load_uniform_grid(data2, ddims, unit_system="code")
    # For this test dataset, code units are SI units
    ds4 = load_uniform_grid(data1, ddims, magnetic_unit=(1.0, "T"), unit_system="code")

    ds0.index
    ds1.index
    ds2.index
    ds3.index
    ds4.index

    dd0 = ds0.all_data()
    dd1 = ds1.all_data()
    dd2 = ds2.all_data()
    dd3 = ds3.all_data()
    dd4 = ds4.all_data()

    assert ds0.fields.gas.magnetic_field_strength.units == "G"
    assert ds1.fields.gas.magnetic_field_strength.units == "G"
    assert ds1.fields.gas.magnetic_field_poloidal.units == "G"
    assert ds1.fields.gas.magnetic_field_toroidal.units == "G"
    assert ds2.fields.gas.magnetic_field_strength.units == "T"
    assert ds2.fields.gas.magnetic_field_poloidal.units == "T"
    assert ds2.fields.gas.magnetic_field_toroidal.units == "T"
    assert ds3.fields.gas.magnetic_field_strength.units == "code_magnetic"
    assert ds3.fields.gas.magnetic_field_poloidal.units == "code_magnetic"
    assert ds3.fields.gas.magnetic_field_toroidal.units == "code_magnetic"
    assert ds4.fields.gas.magnetic_field_strength.units == "code_magnetic"
    assert ds4.fields.gas.magnetic_field_poloidal.units == "code_magnetic"
    assert ds4.fields.gas.magnetic_field_toroidal.units == "code_magnetic"

    emag0 = (
        dd0["gas", "magnetic_field_x"] ** 2
        + dd0["gas", "magnetic_field_y"] ** 2
        + dd0["gas", "magnetic_field_z"] ** 2
    ) / (8.0 * np.pi)
    emag0.convert_to_units("dyne/cm**2")

    emag1 = (
        dd1["gas", "magnetic_field_x"] ** 2
        + dd1["gas", "magnetic_field_y"] ** 2
        + dd1["gas", "magnetic_field_z"] ** 2
    ) / (8.0 * np.pi)
    emag1.convert_to_units("dyne/cm**2")

    emag2 = (
        dd2["gas", "magnetic_field_x"] ** 2
        + dd2["gas", "magnetic_field_y"] ** 2
        + dd2["gas", "magnetic_field_z"] ** 2
    ) / (2.0 * mu_0)
    emag2.convert_to_units("Pa")

    emag3 = (
        dd3["gas", "magnetic_field_x"] ** 2
        + dd3["gas", "magnetic_field_y"] ** 2
        + dd3["gas", "magnetic_field_z"] ** 2
    ) / (8.0 * np.pi)
    emag3.convert_to_units("code_pressure")

    emag4 = (
        dd4["gas", "magnetic_field_x"] ** 2
        + dd4["gas", "magnetic_field_y"] ** 2
        + dd4["gas", "magnetic_field_z"] ** 2
    ) / (2.0 * mu_0)
    emag4.convert_to_units("code_pressure")

    # note that "magnetic_energy_density" and "magnetic_pressure" are aliased

    assert_almost_equal(emag0, dd0["gas", "magnetic_energy_density"])
    assert_almost_equal(emag1, dd1["gas", "magnetic_energy_density"])
    assert_almost_equal(emag2, dd2["gas", "magnetic_energy_density"])
    assert_almost_equal(emag3, dd3["gas", "magnetic_energy_density"])
    assert_almost_equal(emag4, dd4["gas", "magnetic_energy_density"])

    assert str(emag0.units) == str(dd0["gas", "magnetic_energy_density"].units)
    assert str(emag1.units) == str(dd1["gas", "magnetic_energy_density"].units)
    assert str(emag2.units) == str(dd2["gas", "magnetic_energy_density"].units)
    assert str(emag3.units) == str(dd3["gas", "magnetic_energy_density"].units)
    assert str(emag4.units) == str(dd4["gas", "magnetic_energy_density"].units)

    assert_almost_equal(emag1.in_cgs(), emag0.in_cgs())
    assert_almost_equal(emag2.in_cgs(), emag0.in_cgs())
    assert_almost_equal(emag1.in_cgs(), emag2.in_cgs())
    assert_almost_equal(emag1.in_cgs(), emag3.in_cgs())
    assert_almost_equal(emag1.in_cgs(), emag4.in_cgs())



from functools import cached_property

import numpy as np

from yt.utilities.lib.pixelization_routines import pixelize_aitoff, pixelize_cylinder

from .coordinate_handler import (
    CoordinateHandler,
    _get_coord_fields,
    _get_polar_bounds,
    _setup_dummy_cartesian_coords_and_widths,
    _setup_polar_coordinates,
)


class SphericalCoordinateHandler(CoordinateHandler):
    name = "spherical"
    _default_axis_order = ("r", "theta", "phi")

    def __init__(self, ds, ordering=None):
        super().__init__(ds, ordering)
        # Generate
        self.image_units = {}
        self.image_units[self.axis_id["r"]] = (1, 1)
        self.image_units[self.axis_id["theta"]] = (None, None)
        self.image_units[self.axis_id["phi"]] = (None, None)

    def setup_fields(self, registry):
        # Missing implementation for x, y and z coordinates.
        _setup_dummy_cartesian_coords_and_widths(registry, axes=("x", "y", "z"))
        _setup_polar_coordinates(registry, self.axis_id)

        f1, f2 = _get_coord_fields(self.axis_id["phi"], "dimensionless")
        registry.add_field(
            ("index", "dphi"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="dimensionless",
        )

        registry.add_field(
            ("index", "phi"),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="dimensionless",
        )

        def _SphericalVolume(field, data):
            # Here we compute the spherical volume element exactly
            r = data["index", "r"]
            dr = data["index", "dr"]
            theta = data["index", "theta"]
            dtheta = data["index", "dtheta"]
            vol = ((r + 0.5 * dr) ** 3 - (r - 0.5 * dr) ** 3) / 3.0
            vol *= np.cos(theta - 0.5 * dtheta) - np.cos(theta + 0.5 * dtheta)
            vol *= data["index", "dphi"]
            return vol

        registry.add_field(
            ("index", "cell_volume"),
            sampling_type="cell",
            function=_SphericalVolume,
            units="code_length**3",
        )
        registry.alias(("index", "volume"), ("index", "cell_volume"))

        def _path_phi(field, data):
            # Note: this already assumes cell-centered
            return (
                data["index", "r"]
                * data["index", "dphi"]
                * np.sin(data["index", "theta"])
            )

        registry.add_field(
            ("index", "path_element_phi"),
            sampling_type="cell",
            function=_path_phi,
            units="code_length",
        )

    def pixelize(
        self,
        dimension,
        data_source,
        field,
        bounds,
        size,
        antialias=True,
        periodic=True,
        *,
        return_mask=False,
    ):
        self.period
        name = self.axis_name[dimension]
        if name == "r":
            buff, mask = self._ortho_pixelize(
                data_source, field, bounds, size, antialias, dimension, periodic
            )
        elif name in ("theta", "phi"):
            if name == "theta":
                # This is admittedly a very hacky way to resolve a bug
                # it's very likely that the *right* fix would have to
                # be applied upstream of this function, *but* this case
                # has never worked properly so maybe it's still preferable to
                # not having a solution ?
                # see https://github.com/yt-project/yt/pull/3533
                bounds = (*bounds[2:4], *bounds[:2])
            buff, mask = self._cyl_pixelize(
                data_source, field, bounds, size, antialias, dimension
            )
        else:
            raise NotImplementedError

        if return_mask:
            assert mask is None or mask.dtype == bool
            return buff, mask
        else:
            return buff

    def pixelize_line(self, field, start_point, end_point, npoints):
        raise NotImplementedError

    def _ortho_pixelize(
        self, data_source, field, bounds, size, antialias, dim, periodic
    ):
        # use Aitoff projection
        # http://paulbourke.net/geometry/transformationprojection/
        bounds = tuple(_.ndview for _ in self._aitoff_bounds)
        buff, mask = pixelize_aitoff(
            azimuth=data_source["py"],
            dazimuth=data_source["pdy"],
            colatitude=data_source["px"],
            dcolatitude=data_source["pdx"],
            buff_size=size,
            field=data_source[field],
            bounds=bounds,
            input_img=None,
            azimuth_offset=0,
            colatitude_offset=0,
            return_mask=True,
        )
        return buff.T, mask.T

    def _cyl_pixelize(self, data_source, field, bounds, size, antialias, dimension):
        name = self.axis_name[dimension]
        buff = np.full((size[1], size[0]), np.nan, dtype="f8")
        if name == "theta":
            mask = pixelize_cylinder(
                buff,
                data_source["px"],
                data_source["pdx"],
                data_source["py"],
                data_source["pdy"],
                data_source[field],
                bounds,
                return_mask=True,
            )
        elif name == "phi":
            # Note that we feed in buff.T here
            mask = pixelize_cylinder(
                buff.T,
                data_source["px"],
                data_source["pdx"],
                data_source["py"],
                data_source["pdy"],
                data_source[field],
                bounds,
                return_mask=True,
            ).T
        else:
            raise RuntimeError
        return buff, mask

    def convert_from_cartesian(self, coord):
        raise NotImplementedError

    def convert_to_cartesian(self, coord):
        if isinstance(coord, np.ndarray) and len(coord.shape) > 1:
            ri = self.axis_id["r"]
            thetai = self.axis_id["theta"]
            phii = self.axis_id["phi"]
            r = coord[:, ri]
            theta = coord[:, thetai]
            phi = coord[:, phii]
            nc = np.zeros_like(coord)
            # r, theta, phi
            nc[:, ri] = np.cos(phi) * np.sin(theta) * r
            nc[:, thetai] = np.sin(phi) * np.sin(theta) * r
            nc[:, phii] = np.cos(theta) * r
        else:
            r, theta, phi = coord
            nc = (
                np.cos(phi) * np.sin(theta) * r,
                np.sin(phi) * np.sin(theta) * r,
                np.cos(theta) * r,
            )
        return nc

    def convert_to_cylindrical(self, coord):
        raise NotImplementedError

    def convert_from_cylindrical(self, coord):
        raise NotImplementedError

    def convert_to_spherical(self, coord):
        raise NotImplementedError

    def convert_from_spherical(self, coord):
        raise NotImplementedError

    _image_axis_name = None

    @property
    def image_axis_name(self):
        if self._image_axis_name is not None:
            return self._image_axis_name
        # This is the x and y axes labels that get displayed.  For
        # non-Cartesian coordinates, we usually want to override these for
        # Cartesian coordinates, since we transform them.
        rv = {
            self.axis_id["r"]: (
                # these are the Hammer-Aitoff normalized coordinates
                # conventions:
                #   - theta is the colatitude, from 0 to PI
                #   - bartheta is the latitude, from -PI/2 to +PI/2 (bartheta = PI/2 - theta)
                #   - phi is the azimuth, from 0 to 2PI
                #   - lambda is the longitude, from -PI to PI (lambda = phi - PI)
                r"\frac{2\cos(\mathrm{\bar{\theta}})\sin(\lambda/2)}{\sqrt{1 + \cos(\bar{\theta}) \cos(\lambda/2)}}",
                r"\frac{sin(\bar{\theta})}{\sqrt{1 + \cos(\bar{\theta}) \cos(\lambda/2)}}",
                "\\theta",
            ),
            self.axis_id["theta"]: ("x / \\sin(\\theta)", "y / \\sin(\\theta)"),
            self.axis_id["phi"]: ("R", "z"),
        }
        for i in list(rv.keys()):
            rv[self.axis_name[i]] = rv[i]
            rv[self.axis_name[i].capitalize()] = rv[i]
        self._image_axis_name = rv
        return rv

    _x_pairs = (("r", "theta"), ("theta", "r"), ("phi", "r"))
    _y_pairs = (("r", "phi"), ("theta", "phi"), ("phi", "theta"))

    @property
    def period(self):
        return self.ds.domain_width

    @cached_property
    def _poloidal_bounds(self):
        ri = self.axis_id["r"]
        ti = self.axis_id["theta"]
        rmin = self.ds.domain_left_edge[ri]
        rmax = self.ds.domain_right_edge[ri]
        thetamin = self.ds.domain_left_edge[ti]
        thetamax = self.ds.domain_right_edge[ti]
        corners = [
            (rmin, thetamin),
            (rmin, thetamax),
            (rmax, thetamin),
            (rmax, thetamax),
        ]

        def to_poloidal_plane(r, theta):
            # take a r, theta position and return
            # cylindrical R, z coordinates
            R = r * np.sin(theta)
            z = r * np.cos(theta)
            return R, z

        cylindrical_corner_coords = [to_poloidal_plane(*corner) for corner in corners]

        thetamin = thetamin.d
        thetamax = thetamax.d

        Rmin = min(R for R, z in cylindrical_corner_coords)

        if thetamin <= np.pi / 2 <= thetamax:
            Rmax = rmax
        else:
            Rmax = max(R for R, z in cylindrical_corner_coords)

        zmin = min(z for R, z in cylindrical_corner_coords)
        zmax = max(z for R, z in cylindrical_corner_coords)

        return Rmin, Rmax, zmin, zmax

    @cached_property
    def _conic_bounds(self):
        return _get_polar_bounds(self, axes=("r", "phi"))

    @cached_property
    def _aitoff_bounds(self):
        # at the time of writing this function, yt's support for curvilinear
        # coordinates is a bit hacky, as many components of the system still
        # expect to receive coordinates with a length dimension. Ultimately
        # this is not needed but calls for a large refactor.
        ONE = self.ds.quan(1, "code_length")

        # colatitude
        ti = self.axis_id["theta"]
        thetamin = self.ds.domain_left_edge[ti]
        thetamax = self.ds.domain_right_edge[ti]
        # latitude
        latmin = ONE * np.pi / 2 - thetamax
        latmax = ONE * np.pi / 2 - thetamin

        # azimuth
        pi = self.axis_id["phi"]
        phimin = self.ds.domain_left_edge[pi]
        phimax = self.ds.domain_right_edge[pi]
        # longitude
        lonmin = phimin - ONE * np.pi
        lonmax = phimax - ONE * np.pi

        corners = [
            (latmin, lonmin),
            (latmin, lonmax),
            (latmax, lonmin),
            (latmax, lonmax),
        ]

        def aitoff_z(latitude, longitude):
            return np.sqrt(1 + np.cos(latitude) * np.cos(longitude / 2))

        def aitoff_x(latitude, longitude):
            return (
                2
                * np.cos(latitude)
                * np.sin(longitude / 2)
                / aitoff_z(latitude, longitude)
            )

        def aitoff_y(latitude, longitude):
            return np.sin(latitude) / aitoff_z(latitude, longitude)

        def to_aitoff_plane(latitude, longitude):
            return aitoff_x(latitude, longitude), aitoff_y(latitude, longitude)

        aitoff_corner_coords = [to_aitoff_plane(*corner) for corner in corners]

        xmin = ONE * min(x for x, y in aitoff_corner_coords)
        xmax = ONE * max(x for x, y in aitoff_corner_coords)

        # theta is the colatitude
        # What this branch is meant to do is check whether the equator (latitude = 0)
        # is included in the domain.

        if latmin < 0 < latmax:
            xmin = min(xmin, ONE * aitoff_x(0, lonmin))
            xmax = max(xmax, ONE * aitoff_x(0, lonmax))

        # the y direction is more straightforward because aitoff-projected parallels (y)
        # draw a convex shape, while aitoff-projected meridians (x) draw a concave shape
        ymin = ONE * min(y for x, y in aitoff_corner_coords)
        ymax = ONE * max(y for x, y in aitoff_corner_coords)

        return xmin, xmax, ymin, ymax

    def sanitize_center(self, center, axis):
        center, display_center = super().sanitize_center(center, axis)
        name = self.axis_name[axis]
        if name == "r":
            xxmin, xxmax, yymin, yymax = self._aitoff_bounds
            xc = (xxmin + xxmax) / 2
            yc = (yymin + yymax) / 2
            display_center = (0 * xc, xc, yc)
        elif name == "theta":
            xxmin, xxmax, yymin, yymax = self._conic_bounds
            xc = (xxmin + xxmax) / 2
            yc = (yymin + yymax) / 2
            display_center = (xc, 0 * xc, yc)
        elif name == "phi":
            Rmin, Rmax, zmin, zmax = self._poloidal_bounds
            xc = (Rmin + Rmax) / 2
            yc = (zmin + zmax) / 2
            display_center = (xc, yc)
        return center, display_center

    def sanitize_width(self, axis, width, depth):
        name = self.axis_name[axis]
        if width is not None:
            width = super().sanitize_width(axis, width, depth)
        elif name == "r":
            xxmin, xxmax, yymin, yymax = self._aitoff_bounds
            xw = xxmax - xxmin
            yw = yymax - yymin
            width = [xw, yw]
        elif name == "theta":
            # Remember, in spherical coordinates when we cut in theta,
            # we create a conic section
            xxmin, xxmax, yymin, yymax = self._conic_bounds
            xw = xxmax - xxmin
            yw = yymax - yymin
            width = [xw, yw]
        elif name == "phi":
            Rmin, Rmax, zmin, zmax = self._poloidal_bounds
            xw = Rmax - Rmin
            yw = zmax - zmin
            width = [xw, yw]
        return width



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_mass / code_length**3"
vel_units = "code_velocity"
# NOTE: ARTIO uses momentum density.
mom_units = "code_mass / (code_length**2 * code_time)"
en_units = "code_mass*code_velocity**2/code_length**3"


class ARTFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("Density", (rho_units, ["density"], None)),
        ("TotalEnergy", (en_units, ["total_energy_density"], None)),
        ("XMomentumDensity", (mom_units, ["momentum_density_x"], None)),
        ("YMomentumDensity", (mom_units, ["momentum_density_y"], None)),
        ("ZMomentumDensity", (mom_units, ["momentum_density_z"], None)),
        ("Pressure", ("", ["pressure"], None)),  # Unused
        ("Gamma", ("", ["gamma"], None)),
        ("GasEnergy", (en_units, ["thermal_energy_density"], None)),
        ("MetalDensitySNII", (rho_units, ["metal_ii_density"], None)),
        ("MetalDensitySNIa", (rho_units, ["metal_ia_density"], None)),
        ("PotentialNew", ("", ["potential"], None)),
        ("PotentialOld", ("", ["gas_potential"], None)),
    )

    known_particle_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_index", ("", [], None)),
        ("particle_species", ("", ["particle_type"], None)),
        ("particle_creation_time", ("Gyr", [], None)),
        ("particle_mass_initial", ("code_mass", [], None)),
        ("particle_metallicity1", ("", [], None)),
        ("particle_metallicity2", ("", [], None)),
    )

    def setup_fluid_fields(self):
        unit_system = self.ds.unit_system

        def _temperature(field, data):
            r0 = data.ds.parameters["boxh"] / data.ds.parameters["ng"]
            tr = data.ds.quan(3.03e5 * r0**2, "K/code_velocity**2")
            tr *= data.ds.parameters["wmu"] * data.ds.parameters["Om0"]
            tr *= data.ds.parameters["gamma"] - 1.0
            tr /= data.ds.parameters["aexpn"] ** 2
            return tr * data["art", "GasEnergy"] / data["art", "Density"]

        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=unit_system["temperature"],
        )

        def _get_vel(axis):
            def velocity(field, data):
                return data["gas", f"momentum_density_{axis}"] / data["gas", "density"]

            return velocity

        for ax in "xyz":
            self.add_field(
                ("gas", f"velocity_{ax}"),
                sampling_type="cell",
                function=_get_vel(ax),
                units=unit_system["velocity"],
            )

        def _momentum_magnitude(field, data):
            tr = (
                data["gas", "momentum_density_x"] ** 2
                + data["gas", "momentum_density_y"] ** 2
                + data["gas", "momentum_density_z"] ** 2
            ) ** 0.5
            tr *= data["index", "cell_volume"].in_units("cm**3")
            return tr

        self.add_field(
            ("gas", "momentum_magnitude"),
            sampling_type="cell",
            function=_momentum_magnitude,
            units=unit_system["momentum"],
        )

        def _velocity_magnitude(field, data):
            tr = data["gas", "momentum_magnitude"]
            tr /= data["gas", "cell_mass"]
            return tr

        self.add_field(
            ("gas", "velocity_magnitude"),
            sampling_type="cell",
            function=_velocity_magnitude,
            units=unit_system["velocity"],
        )

        def _metal_density(field, data):
            tr = data["gas", "metal_ia_density"]
            tr += data["gas", "metal_ii_density"]
            return tr

        self.add_field(
            ("gas", "metal_density"),
            sampling_type="cell",
            function=_metal_density,
            units=unit_system["density"],
        )

        def _metal_mass_fraction(field, data):
            tr = data["gas", "metal_density"]
            tr /= data["gas", "density"]
            return tr

        self.add_field(
            ("gas", "metal_mass_fraction"),
            sampling_type="cell",
            function=_metal_mass_fraction,
            units="",
        )

        def _H_mass_fraction(field, data):
            tr = 1.0 - data.ds.parameters["Y_p"] - data["gas", "metal_mass_fraction"]
            return tr

        self.add_field(
            ("gas", "H_mass_fraction"),
            sampling_type="cell",
            function=_H_mass_fraction,
            units="",
        )

        def _metallicity(field, data):
            tr = data["gas", "metal_mass_fraction"]
            tr /= data["gas", "H_mass_fraction"]
            return tr

        self.add_field(
            ("gas", "metallicity"),
            sampling_type="cell",
            function=_metallicity,
            units="",
        )

        atoms = [
            "C",
            "N",
            "O",
            "F",
            "Ne",
            "Na",
            "Mg",
            "Al",
            "Si",
            "P",
            "S",
            "Cl",
            "Ar",
            "K",
            "Ca",
            "Sc",
            "Ti",
            "V",
            "Cr",
            "Mn",
            "Fe",
            "Co",
            "Ni",
            "Cu",
            "Zn",
        ]

        def _specific_metal_density_function(atom):
            def _specific_metal_density(field, data):
                nucleus_densityIa = (
                    data["gas", "metal_ia_density"] * SNIa_abundance[atom]
                )
                nucleus_densityII = (
                    data["gas", "metal_ii_density"] * SNII_abundance[atom]
                )
                return nucleus_densityIa + nucleus_densityII

            return _specific_metal_density

        for atom in atoms:
            self.add_field(
                ("gas", f"{atom}_nuclei_mass_density"),
                sampling_type="cell",
                function=_specific_metal_density_function(atom),
                units=unit_system["density"],
            )


# based on Iwamoto et al 1999
# mass fraction of each atom in SNIa metal
SNIa_abundance = {
    "H": 0.00e00,
    "He": 0.00e00,
    "C": 3.52e-02,
    "N": 8.47e-07,
    "O": 1.04e-01,
    "F": 4.14e-10,
    "Ne": 3.30e-03,
    "Na": 4.61e-05,
    "Mg": 6.25e-03,
    "Al": 7.19e-04,
    "Si": 1.14e-01,
    "P": 2.60e-04,
    "S": 6.35e-02,
    "Cl": 1.27e-04,
    "Ar": 1.14e-02,
    "K": 5.72e-05,
    "Ca": 8.71e-03,
    "Sc": 1.61e-07,
    "Ti": 2.50e-04,
    "V": 5.46e-05,
    "Cr": 6.19e-03,
    "Mn": 6.47e-03,
    "Fe": 5.46e-01,
    "Co": 7.59e-04,
    "Ni": 9.17e-02,
    "Cu": 2.19e-06,
    "Zn": 2.06e-05,
}

# mass fraction of each atom in SNII metal
SNII_abundance = {
    "H": 0.00e00,
    "He": 0.00e00,
    "C": 3.12e-02,
    "N": 6.15e-04,
    "O": 7.11e-01,
    "F": 4.57e-10,
    "Ne": 9.12e-02,
    "Na": 2.56e-03,
    "Mg": 4.84e-02,
    "Al": 5.83e-03,
    "Si": 4.81e-02,
    "P": 4.77e-04,
    "S": 1.62e-02,
    "Cl": 4.72e-05,
    "Ar": 3.15e-03,
    "K": 2.65e-05,
    "Ca": 2.31e-03,
    "Sc": 9.02e-08,
    "Ti": 5.18e-05,
    "V": 3.94e-06,
    "Cr": 5.18e-04,
    "Mn": 1.52e-04,
    "Fe": 3.58e-02,
    "Co": 2.86e-05,
    "Ni": 2.35e-03,
    "Cu": 4.90e-07,
    "Zn": 7.46e-06,
}



from operator import attrgetter

import numpy as np

from yt.utilities.io_handler import BaseParticleIOHandler


class IOHandlerAHFHalos(BaseParticleIOHandler):
    _particle_reader = False
    _dataset_type = "ahf"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    def _read_particle_coords(self, chunks, ptf):
        # This needs to *yield* a series of tuples of (ptype, (x, y, z), hsml).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.

        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        for data_file in self._sorted_chunk_iterator(chunks):
            pos = data_file._get_particle_positions("halos")
            x, y, z = (pos[:, i] for i in range(3))
            yield "halos", (x, y, z), 0.0

    def _yield_coordinates(self, data_file):
        halos = data_file.read_data(usecols=["Xc", "Yc", "Zc"])
        x = halos["Xc"].astype("float64")
        y = halos["Yc"].astype("float64")
        z = halos["Zc"].astype("float64")
        yield "halos", np.asarray((x, y, z)).T

    def _read_particle_fields(self, chunks, ptf, selector):
        # This gets called after the arrays have been allocated.  It needs to
        # yield ((ptype, field), data) where data is the masked results of
        # reading ptype, field and applying the selector to the data read in.
        # Selector objects have a .select_points(x,y,z) that returns a mask, so
        # you need to do your masking here.
        # Only support halo reading for now.
        assert len(ptf) == 1
        assert list(ptf.keys())[0] == "halos"
        for data_file in self._sorted_chunk_iterator(chunks):
            si, ei = data_file.start, data_file.end
            cols = []
            for field_list in ptf.values():
                cols.extend(field_list)
            cols = list(set(cols))
            halos = data_file.read_data(usecols=cols)
            pos = data_file._get_particle_positions("halos")
            x, y, z = (pos[:, i] for i in range(3))
            yield "halos", (x, y, z)
            mask = selector.select_points(x, y, z, 0.0)
            del x, y, z
            if mask is None:
                continue
            for ptype, field_list in sorted(ptf.items()):
                for field in field_list:
                    data = halos[field][si:ei][mask].astype("float64")
                    yield (ptype, field), data

    def _count_particles(self, data_file):
        halos = data_file.read_data(usecols=["ID"])
        nhalos = len(halos["ID"])
        si, ei = data_file.start, data_file.end
        if None not in (si, ei):
            nhalos = np.clip(nhalos - si, 0, ei - si)
        return {"halos": nhalos}

    def _identify_fields(self, data_file):
        fields = [("halos", f) for f in data_file.col_names]
        return fields, {}

    def _sorted_chunk_iterator(self, chunks):
        # yield from sorted list of data_files
        data_files = self._get_data_files(chunks)
        yield from sorted(data_files, key=attrgetter("filename"))



import os
import warnings
import weakref

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.fields.magnetic_field import get_magnetic_normalization
from yt.funcs import mylog, setdefaultattr
from yt.geometry.api import Geometry
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.chemical_formulas import compute_mu
from yt.utilities.file_handler import HDF5FileHandler

from .fields import ParthenonFieldInfo

_geom_map = {
    "UniformCartesian": Geometry.CARTESIAN,
    "UniformCylindrical": Geometry.CYLINDRICAL,
    "UniformSpherical": Geometry.SPHERICAL,
}

# fmt: off
_cis = np.array(
    [
        [0, 0, 0],
        [0, 0, 1],
        [0, 1, 0],
        [0, 1, 1],
        [1, 0, 0],
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1],
    ],
    dtype="int64",
)
# fmt: on


class ParthenonGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz , at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        LE, RE = self.index.grid_left_edge[id, :], self.index.grid_right_edge[id, :]
        self.dds = self.ds.arr((RE - LE) / self.ActiveDimensions, "code_length")
        self.field_data["dx"], self.field_data["dy"], self.field_data["dz"] = self.dds

    def retrieve_ghost_zones(self, n_zones, fields, all_levels=False, smoothed=False):
        if smoothed:
            warnings.warn(
                "ghost-zones interpolation/smoothing is not "
                "currently supported for Parthenon data.",
                category=RuntimeWarning,
                stacklevel=2,
            )
            smoothed = False
        return super().retrieve_ghost_zones(
            n_zones, fields, all_levels=all_levels, smoothed=smoothed
        )


class ParthenonHierarchy(GridIndex):
    grid = ParthenonGrid
    _dataset_type = "parthenon"
    _data_file = None

    def __init__(self, ds, dataset_type="parthenon"):
        self.dataset = weakref.proxy(ds)
        self.directory = os.path.dirname(self.dataset.filename)
        self.dataset_type = dataset_type
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.filename
        self._handle = ds._handle
        GridIndex.__init__(self, ds, dataset_type)

    def _detect_output_fields(self):
        self.field_list = [("parthenon", k) for k in self.dataset._field_map]

    def _count_grids(self):
        self.num_grids = self._handle["Info"].attrs["NumMeshBlocks"]

    def _parse_index(self):
        num_grids = self._handle["Info"].attrs["NumMeshBlocks"]

        # TODO: In an unlikely case this would use too much memory, implement
        #       chunked read along 1 dim
        x = self._handle["Locations"]["x"][:, :]
        y = self._handle["Locations"]["y"][:, :]
        z = self._handle["Locations"]["z"][:, :]
        mesh_block_size = self._handle["Info"].attrs["MeshBlockSize"]

        self.grids = np.empty(self.num_grids, dtype="object")
        levels = self._handle["Levels"][:]
        for i in range(num_grids):
            self.grid_left_edge[i] = np.array(
                [x[i, 0], y[i, 0], z[i, 0]], dtype="float64"
            )
            self.grid_right_edge[i] = np.array(
                [x[i, -1], y[i, -1], z[i, -1]], dtype="float64"
            )
            self.grid_dimensions[i] = mesh_block_size
            self.grids[i] = self.grid(i, self, levels[i])

        if self.dataset.dimensionality <= 2:
            self.grid_right_edge[:, 2] = self.dataset.domain_right_edge[2]
        if self.dataset.dimensionality == 1:
            self.grid_right_edge[:, 1:] = self.dataset.domain_right_edge[1:]
        self.grid_particle_count = np.zeros([self.num_grids, 1], dtype="int64")

    def _populate_grid_objects(self):
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()
        self.max_level = self._handle["Info"].attrs["MaxLevel"]


class ParthenonDataset(Dataset):
    _load_requirements = ["h5py"]
    _field_info_class = ParthenonFieldInfo
    _dataset_type = "parthenon"
    _index_class = ParthenonHierarchy

    def __init__(
        self,
        filename,
        dataset_type="parthenon",
        storage_filename=None,
        parameters=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
        magnetic_normalization="gaussian",
    ):
        self.fluid_types += ("parthenon",)
        if parameters is None:
            parameters = {}
        self.specified_parameters = parameters
        if units_override is None:
            units_override = {}
        self._handle = HDF5FileHandler(filename)
        xrat = self._handle["Info"].attrs["RootGridDomain"][2]
        yrat = self._handle["Info"].attrs["RootGridDomain"][5]
        zrat = self._handle["Info"].attrs["RootGridDomain"][8]
        if xrat != 1.0 or yrat != 1.0 or zrat != 1.0:
            raise NotImplementedError(
                "Logarithmic grids not yet supported/tested in Parthenon frontend."
            )

        self._magnetic_factor = get_magnetic_normalization(magnetic_normalization)

        self.geometry = _geom_map[self._handle["Info"].attrs["Coordinates"]]

        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        if storage_filename is None:
            storage_filename = self.basename + ".yt"
        self.storage_filename = storage_filename

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the
        parameter file
        """
        for unit, cgs in [
            ("length", "cm"),
            ("time", "s"),
            ("mass", "g"),
        ]:
            unit_param = f"Hydro/code_{unit}_cgs"
            # Use units, if provided in output
            if unit_param in self.parameters:
                setdefaultattr(
                    self, f"{unit}_unit", self.quan(self.parameters[unit_param], cgs)
                )
            # otherwise use code = cgs
            else:
                mylog.warning(f"Assuming 1.0 code_{unit} = 1.0 {cgs}")
                setdefaultattr(self, f"{unit}_unit", self.quan(1.0, cgs))

        self.magnetic_unit = np.sqrt(
            self._magnetic_factor
            * self.mass_unit
            / (self.time_unit**2 * self.length_unit)
        )
        self.magnetic_unit.convert_to_units("gauss")
        self.velocity_unit = self.length_unit / self.time_unit

    def _parse_parameter_file(self):
        self.parameters.update(self.specified_parameters)
        for key, val in self._handle["Params"].attrs.items():
            if key in self.parameters.keys():
                mylog.warning(
                    f"Overriding existing {key!r} key in ds.parameters from data 'Params'"
                )
            self.parameters[key] = val

        xmin, xmax = self._handle["Info"].attrs["RootGridDomain"][0:2]
        ymin, ymax = self._handle["Info"].attrs["RootGridDomain"][3:5]
        zmin, zmax = self._handle["Info"].attrs["RootGridDomain"][6:8]

        self.domain_left_edge = np.array([xmin, ymin, zmin], dtype="float64")
        self.domain_right_edge = np.array([xmax, ymax, zmax], dtype="float64")

        self.domain_width = self.domain_right_edge - self.domain_left_edge
        self.domain_dimensions = self._handle["Info"].attrs["RootGridSize"]

        self._field_map = {}
        k = 0

        dnames = self._handle["Info"].attrs["OutputDatasetNames"]
        num_components = self._handle["Info"].attrs["NumComponents"]

        if "OutputFormatVersion" in self._handle["Info"].attrs.keys():
            self.output_format_version = self._handle["Info"].attrs[
                "OutputFormatVersion"
            ]
        else:
            raise NotImplementedError("Could not determine OutputFormatVersion.")

        # For a single variable, we need to convert it to a list for the following
        # zip to work.
        if isinstance(num_components, np.uint64):
            dnames = (dnames,)
            num_components = (num_components,)

        component_name_offset = 0
        for dname, num_component in zip(dnames, num_components, strict=False):
            for j in range(num_component):
                fname = self._handle["Info"].attrs["ComponentNames"][
                    j + component_name_offset
                ]
                self._field_map[fname] = (dname, j)
                k += 1
            component_name_offset = int(component_name_offset + num_component)

        self.refine_by = 2
        dimensionality = 3
        if self.domain_dimensions[2] == 1:
            dimensionality = 2
        if self.domain_dimensions[1] == 1:
            dimensionality = 1
        self.dimensionality = dimensionality
        self.current_time = self._handle["Info"].attrs["Time"]
        self.num_ghost_zones = 0
        self.field_ordering = "fortran"
        self.boundary_conditions = [1] * 6
        self.cosmological_simulation = False

        if "periodicity" in self.parameters:
            self._periodicity = tuple(self.parameters["periodicity"])
        else:
            boundary_conditions = self._handle["Info"].attrs["BoundaryConditions"]

            inner_bcs = boundary_conditions[::2]
            # outer_bcs = boundary_conditions[1::2]
            ##Check self consistency
            # for inner_bc,outer_bc in zip(inner_bcs,outer_bcs):
            #    if( (inner_bc == "periodicity" or outer_bc == "periodic") and inner_bc != outer_bc ):
            #        raise Exception("Inconsistent periodicity in boundary conditions")

            self._periodicity = tuple(bc == "periodic" for bc in inner_bcs)

        if "gamma" in self.parameters:
            self.gamma = float(self.parameters["gamma"])
        elif "Hydro/AdiabaticIndex" in self.parameters:
            self.gamma = self.parameters["Hydro/AdiabaticIndex"]
        else:
            mylog.warning(
                "Adiabatic index gamma could not be determined. Falling back to 5/3."
            )
            self.gamma = 5.0 / 3.0

        if "mu" in self.parameters:
            self.mu = self.parameters["mu"]
        elif "Hydro/mu" in self.parameters:
            self.mu = self.parameters["Hydro/mu"]
        # Legacy He_mass_fraction parameter implemented in AthenaPK
        elif "Hydro/He_mass_fraction" in self.parameters:
            He_mass_fraction = self.parameters["Hydro/He_mass_fraction"]
            self.mu = 1 / (He_mass_fraction * 3.0 / 4.0 + (1 - He_mass_fraction) * 2)
        # Fallback to primorial gas composition (and show warning)
        else:
            mylog.warning(
                "Plasma composition could not be determined in data file. Falling back to fully ionized primodial composition."
            )
            self.mu = self.parameters.get("mu", compute_mu(self.default_species_fields))

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False
        return filename.endswith((".phdf", ".rhdf"))

    @property
    def _skip_cache(self):
        return True

    def __str__(self):
        return self.basename.rsplit(".", 1)[0]



import numpy as np

from yt.utilities.file_handler import NetCDF4FileHandler
from yt.utilities.io_handler import BaseIOHandler


class IOHandlerExodusII(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "exodus_ii"
    _INDEX_OFFSET = 1

    def __init__(self, ds):
        self.filename = ds.index_filename
        exodus_ii_handler = NetCDF4FileHandler(self.filename)
        self.handler = exodus_ii_handler
        super().__init__(ds)
        self.node_fields = ds._get_nod_names()
        self.elem_fields = ds._get_elem_names()

    def _read_particle_coords(self, chunks, ptf):
        pass

    def _read_particle_fields(self, chunks, ptf, selector):
        pass

    def _read_fluid_selection(self, chunks, selector, fields, size):
        # This needs to allocate a set of arrays inside a dictionary, where the
        # keys are the (ftype, fname) tuples and the values are arrays that
        # have been masked using whatever selector method is appropriate.  The
        # dict gets returned at the end and it should be flat, with selected
        # data.  Note that if you're reading grid data, you might need to
        # special-case a grid selector object.
        with self.handler.open_ds() as ds:
            chunks = list(chunks)
            rv = {}
            for field in fields:
                ftype, fname = field
                if ftype == "all":
                    ci = np.concatenate(
                        [
                            mesh.connectivity_indices - self._INDEX_OFFSET
                            for mesh in self.ds.index.mesh_union
                        ]
                    )
                else:
                    ci = ds.variables[ftype][:] - self._INDEX_OFFSET
                num_elem = ci.shape[0]
                if fname in self.node_fields:
                    nodes_per_element = ci.shape[1]
                    rv[field] = np.zeros((num_elem, nodes_per_element), dtype="float64")
                elif fname in self.elem_fields:
                    rv[field] = np.zeros(num_elem, dtype="float64")
            for field in fields:
                ind = 0
                ftype, fname = field
                if ftype == "all":
                    mesh_ids = [mesh.mesh_id + 1 for mesh in self.ds.index.mesh_union]
                    objs = list(self.ds.index.mesh_union)
                else:
                    mesh_ids = [int(ftype.replace("connect", ""))]
                    chunk = chunks[mesh_ids[0] - 1]
                    objs = chunk.objs
                if fname in self.node_fields:
                    field_ind = self.node_fields.index(fname)
                    fdata = ds.variables["vals_nod_var%d" % (field_ind + 1)]
                    for g in objs:
                        ci = g.connectivity_indices - self._INDEX_OFFSET
                        data = fdata[self.ds.step][ci]
                        ind += g.select(selector, data, rv[field], ind)  # caches
                if fname in self.elem_fields:
                    field_ind = self.elem_fields.index(fname)
                    for g, mesh_id in zip(objs, mesh_ids, strict=True):
                        fdata = ds.variables[
                            "vals_elem_var%deb%s" % (field_ind + 1, mesh_id)
                        ][:]
                        data = fdata[self.ds.step, :]
                        ind += g.select(selector, data, rv[field], ind)  # caches
                rv[field] = rv[field][:ind]
            return rv

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk, and is only used for
        # caching.
        pass



import numpy as np
from numpy.testing import assert_equal

from yt.testing import assert_rel_equal
from yt.utilities.lib.ragged_arrays import index_unop

operations = ((np.sum, "sum"), (np.prod, "prod"), (np.max, "max"), (np.min, "min"))
dtypes = (
    (-1e8, 1e8, "float32"),
    (-1e8, 1e8, "float64"),
    (-10000, 10000, "int32"),
    (-100000000, 100000000, "int64"),
)


def test_index_unop():
    np.random.seed(0x4D3D3D3)
    indices = np.arange(1000, dtype="int64")
    np.random.shuffle(indices)
    sizes = np.array([200, 50, 50, 100, 32, 32, 32, 32, 32, 64, 376], dtype="int64")
    for mi, ma, dtype in dtypes:
        for op, operation in operations:
            # Create a random set of values
            values = np.random.random(1000)
            if operation != "prod":
                values = values * ma + (ma - mi)
            if operation == "prod" and dtype.startswith("int"):
                values = values.astype(dtype)
                values[values != 0] = 1
                values[values == 0] = -1
            values = values.astype(dtype)
            out_values = index_unop(values, indices, sizes, operation)
            i = 0
            for j, v in enumerate(sizes):
                arr = values[indices[i : i + v]]
                if dtype == "float32":
                    # Numpy 1.9.1 changes the accumulator type to promote
                    assert_rel_equal(op(arr), out_values[j], 6)
                elif dtype == "float64":
                    # Numpy 1.9.1 changes the accumulator type to promote
                    assert_rel_equal(op(arr), out_values[j], 12)
                else:
                    assert_equal(op(arr), out_values[j])
                i += v



import numpy as np

nar = np.array

counter = 0


def recursive_all_clumps(clump, list, level, parentnumber):
    """A recursive function to flatten the index in *clump*.
    Not to be called directly: please call return_all_clumps, below."""
    global counter
    counter += 1
    clump.number = counter
    clump.parentnumber = parentnumber
    counter += 1
    list.append(clump)
    clump.level = level
    if clump.children is not None:
        for child in clump.children:
            recursive_all_clumps(child, list, level + 1, clump.number)
    return list


def return_all_clumps(clump):
    """Flatten the index defined by *clump*.
    Additionally adds three variables to the clump:
    level        = depth of index
    number       = index of clump in the final array
    parentnumber = index of this clumps parent

    """
    global counter
    counter = 0
    list = []
    level = 0
    clump.level = level
    parentnumber = -1
    recursive_all_clumps(clump, list, level, parentnumber)
    return list


def return_bottom_clumps(clump, dbg=0):
    """
    Recursively return clumps at the bottom of the index.
    This gives a list of clumps similar to what one would get from a CLUMPFIND routine
    """
    global counter
    counter = 0
    list = []
    level = 0
    recursive_bottom_clumps(clump, list, dbg, level)
    return list


def recursive_bottom_clumps(clump, clump_list, dbg=0, level=0):
    """Loops over a list of clumps (clumps) and fills clump_list with the bottom most.
    Recursive. Prints the level and the number of cores to the screen."""

    global counter

    if (clump.children is None) or (len(clump.children) == 0):
        counter += 1
        clump_list.append(clump)
    else:
        for child in clump.children:
            recursive_bottom_clumps(child, clump_list, dbg=dbg, level=level + 1)


def clump_list_sort(clump_list):
    """Returns a copy of clump_list, sorted by ascending minimum density.  This
    eliminates overlap when passing to
    yt.visualization.plot_modification.ClumpContourCallback"""
    minDensity = [c["gas", "density"].min() for c in clump_list]

    args = np.argsort(minDensity)
    list = nar(clump_list)[args]
    reverse = range(list.size - 1, -1, -1)
    return list[reverse]



from yt.loaders import load
from yt.sample_data.api import _get_test_data_dir_path
from yt.testing import assert_array_equal, requires_file


@requires_file("TNGHalo/halo_59.hdf5")
def test_ewah_write_load(tmp_path):
    mock_file = tmp_path / "halo_59.hdf5"
    mock_file.symlink_to(_get_test_data_dir_path() / "TNGHalo" / "halo_59.hdf5")

    masses = []
    opts = [
        (None, True, (6, 2, 4), False),
        (None, True, (6, 2, 4), True),
        ((5, 3), False, (5, 3, 3), False),
        ((5, 3), False, (5, 3, 3), True),
    ]

    for opt in opts:
        ds = load(mock_file, index_order=opt[0])
        _, c = ds.find_max(("gas", "density"))
        assert ds.index.pii.mutable_index is opt[1]
        assert ds.index.pii.order1 == opt[2][0]
        assert ds.index.pii.order2_orig == opt[2][1]
        assert ds.index.pii.order2 == opt[2][2]
        assert ds.index.pii._is_loaded is opt[3]
        c += ds.quan(0.5, "Mpc")
        sp = ds.sphere(c, (20.0, "kpc"))
        mass = sp.sum(("gas", "mass"))
        masses.append(mass)

    assert_array_equal(masses[0], masses[1:])



import numpy as np

from yt.frontends.enzo_e.misc import get_particle_mass_correction, nested_dict_get
from yt.utilities.exceptions import YTException
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.on_demand_imports import _h5py as h5py


class EnzoEIOHandler(BaseIOHandler):
    _dataset_type = "enzo_e"
    _base = slice(None)
    _field_dtype = "float64"
    _sep = "_"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # precompute the indexing specifying each field's active zone
        # -> this assumes that each field in Enzo-E shares the same number of
        #    ghost-zones. Technically, Enzo-E allows each field to have a
        #    different number of ghost zones (but this feature isn't currently
        #    used & it currently doesn't record this information)
        # -> our usage of a negative stop value ensures compatability with
        #    both cell-centered and face-centered fields
        self._activezone_idx = tuple(
            slice(num_zones, -num_zones) if num_zones > 0 else slice(None)
            for num_zones in self.ds.ghost_zones[: self.ds.dimensionality]
        )

        # Determine if particle masses are actually masses or densities.
        if self.ds.parameters["version"] is not None:
            # they're masses for enzo-e versions that record a version string
            mass_flag = True
        else:
            # in earlier versions: query the existence of the "mass_is_mass"
            # particle parameter
            mass_flag = nested_dict_get(
                self.ds.parameters, ("Particle", "mass_is_mass"), default=None
            )
        # the historic approach for initializing the value of "mass_is_mass"
        # was unsound (and could yield a random value). Thus we should only
        # check for the parameter's existence and not its value
        self._particle_mass_is_mass = mass_flag is not None

    def _read_field_names(self, grid):
        if grid.filename is None:
            return []
        f = h5py.File(grid.filename, mode="r")
        try:
            group = f[grid.block_name]
        except KeyError as e:
            raise YTException(
                message=f"Grid {grid.block_name} is missing from data file {grid.filename}.",
                ds=self.ds,
            ) from e
        fields = []
        ptypes = set()
        dtypes = set()
        # keep one field for each particle type so we can count later
        sample_pfields = {}
        for name, v in group.items():
            if not hasattr(v, "shape") or v.dtype == "O":
                continue
            # mesh fields are "field <name>"
            if name.startswith("field"):
                _, fname = name.split(self._sep, 1)
                fields.append(("enzoe", fname))
                dtypes.add(v.dtype)
            # particle fields are "particle <type> <name>"
            else:
                _, ftype, fname = name.split(self._sep, 2)
                fields.append((ftype, fname))
                ptypes.add(ftype)
                dtypes.add(v.dtype)
                if ftype not in sample_pfields:
                    sample_pfields[ftype] = fname
        self.sample_pfields = sample_pfields

        if len(dtypes) == 1:
            # Now, if everything we saw was the same dtype, we can go ahead and
            # set it here.  We do this because it is a HUGE savings for 32 bit
            # floats, since our numpy copying/casting is way faster than
            # h5py's, for some reason I don't understand.  This does *not* need
            # to be correct -- it will get fixed later -- it just needs to be
            # okay for now.
            self._field_dtype = list(dtypes)[0]
        f.close()
        return fields, ptypes

    def _read_particle_coords(self, chunks, ptf):
        yield from (
            (ptype, xyz, 0.0)
            for ptype, xyz in self._read_particle_fields(chunks, ptf, None)
        )

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        dc = self.ds.domain_center.in_units("code_length").d
        for chunk in chunks:  # These should be organized by grid filename
            f = None
            for g in chunk.objs:
                if g.filename is None:
                    continue
                if f is None:
                    f = h5py.File(g.filename, mode="r")
                if g.particle_count is None:
                    fnstr = "{}/{}".format(
                        g.block_name,
                        self._sep.join(["particle", "%s", "%s"]),
                    )
                    g.particle_count = {
                        ptype: f.get(fnstr % (ptype, self.sample_pfields[ptype])).size
                        for ptype in self.sample_pfields
                    }
                    g.total_particles = sum(g.particle_count.values())
                if g.total_particles == 0:
                    continue
                group = f.get(g.block_name)
                for ptype, field_list in sorted(ptf.items()):
                    pn = self._sep.join(["particle", ptype, "%s"])
                    if g.particle_count[ptype] == 0:
                        continue
                    coords = tuple(
                        np.asarray(group.get(pn % ax)[()], dtype="=f8")
                        for ax in "xyz"[: self.ds.dimensionality]
                    )
                    for i in range(self.ds.dimensionality, 3):
                        coords += (
                            dc[i] * np.ones(g.particle_count[ptype], dtype="f8"),
                        )
                    if selector is None:
                        # This only ever happens if the call is made from
                        # _read_particle_coords.
                        yield ptype, coords
                        continue
                    coords += (0.0,)
                    mask = selector.select_points(*coords)
                    if mask is None:
                        continue
                    for field in field_list:
                        data = np.asarray(group.get(pn % field)[()], "=f8")
                        if field == "mass" and not self._particle_mass_is_mass:
                            data[mask] *= get_particle_mass_correction(self.ds)
                        yield (ptype, field), data[mask]
            if f:
                f.close()

    def io_iter(self, chunks, fields):
        for chunk in chunks:
            fid = None
            filename = -1
            for obj in chunk.objs:
                if obj.filename is None:
                    continue
                if obj.filename != filename:
                    # Note one really important thing here: even if we do
                    # implement LRU caching in the _read_obj_field function,
                    # we'll still be doing file opening and whatnot.  This is a
                    # problem, but one we can return to.
                    if fid is not None:
                        fid.close()
                    fid = h5py.h5f.open(
                        obj.filename.encode("latin-1"), h5py.h5f.ACC_RDONLY
                    )
                    filename = obj.filename
                for field in fields:
                    grid_dim = self.ds.grid_dimensions
                    nodal_flag = self.ds.field_info[field].nodal_flag
                    dims = (
                        grid_dim[: self.ds.dimensionality][::-1]
                        + nodal_flag[: self.ds.dimensionality][::-1]
                    )
                    data = np.empty(dims, dtype=self._field_dtype)
                    yield field, obj, self._read_obj_field(obj, field, (fid, data))
        if fid is not None:
            fid.close()

    def _read_obj_field(self, obj, field, fid_data):
        if fid_data is None:
            fid_data = (None, None)
        fid, rdata = fid_data
        if fid is None:
            close = True
            fid = h5py.h5f.open(obj.filename.encode("latin-1"), h5py.h5f.ACC_RDONLY)
        else:
            close = False
        ftype, fname = field
        node = f"/{obj.block_name}/field{self._sep}{fname}"
        dg = h5py.h5d.open(fid, node.encode("latin-1"))
        if rdata is None:
            rdata = np.empty(
                self.ds.grid_dimensions[: self.ds.dimensionality][::-1],
                dtype=self._field_dtype,
            )
        dg.read(h5py.h5s.ALL, h5py.h5s.ALL, rdata)
        if close:
            fid.close()
        data = rdata[self._activezone_idx].T
        if self.ds.dimensionality < 3:
            nshape = data.shape + (1,) * (3 - self.ds.dimensionality)
            data = np.reshape(data, nshape)
        return data



from yt.fields.xray_emission_fields import add_xray_emissivity_field
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    ProjectionValuesTest,
    can_run_ds,
    data_dir_load,
    requires_ds,
)


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def check_xray_fields(ds_fn, fields):
    if not can_run_ds(ds_fn):
        return
    dso = [None, ("sphere", ("m", (0.1, "unitary")))]
    for field in fields:
        for dobj_name in dso:
            for axis in [0, 1, 2]:
                yield ProjectionValuesTest(ds_fn, axis, field, None, dobj_name)
            yield FieldValuesTest(ds_fn, field, dobj_name)


sloshing = "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0300"
d9p = "D9p_500/10MpcBox_HartGal_csf_a0.500.d"


@requires_ds(sloshing, big_data=True)
def test_sloshing_apec():
    ds = data_dir_load(sloshing, kwargs={"default_species_fields": "ionized"})
    fields = add_xray_emissivity_field(ds, 0.5, 7.0, table_type="apec", metallicity=0.3)
    for test in check_xray_fields(ds, fields):
        test_sloshing_apec.__name__ = test.description
        yield test


@requires_ds(d9p, big_data=True)
def test_d9p_cloudy():
    ds = data_dir_load(d9p, kwargs={"default_species_fields": "ionized"})
    fields = add_xray_emissivity_field(
        ds,
        0.5,
        2.0,
        redshift=ds.current_redshift,
        table_type="cloudy",
        cosmology=ds.cosmology,
        metallicity=("gas", "metallicity"),
    )
    for test in check_xray_fields(ds, fields):
        test.suffix = "current_redshift"
        test_d9p_cloudy.__name__ = test.description + test.suffix
        yield test


@requires_ds(d9p, big_data=True)
def test_d9p_cloudy_local():
    ds = data_dir_load(d9p, kwargs={"default_species_fields": "ionized"})
    fields = add_xray_emissivity_field(
        ds,
        0.5,
        2.0,
        dist=(1.0, "Mpc"),
        table_type="cloudy",
        metallicity=("gas", "metallicity"),
    )
    for test in check_xray_fields(ds, fields):
        test.suffix = "dist_1Mpc"
        test_d9p_cloudy_local.__name__ = test.description + test.suffix
        yield test



import numpy as np
from numpy.testing import assert_equal

from yt.testing import fake_amr_ds, fake_random_ds


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_mean_sum_integrate():
    for nprocs in [-1, 1, 2, 16]:
        if nprocs == -1:
            ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), particles=20)
        else:
            ds = fake_random_ds(
                32, nprocs=nprocs, fields=("density",), units=("g/cm**3",), particles=20
            )
        ad = ds.all_data()

        # Sums
        q = ad.sum(("gas", "density"))

        q1 = ad.quantities.total_quantity(("gas", "density"))

        assert_equal(q, q1)

        q = ad.sum(("all", "particle_ones"))

        q1 = ad.quantities.total_quantity(("all", "particle_ones"))

        assert_equal(q, q1)

        # Weighted Averages
        w = ad.mean(("gas", "density"))

        w1 = ad.quantities.weighted_average_quantity(
            ("gas", "density"), ("index", "ones")
        )

        assert_equal(w, w1)

        w = ad.mean(("gas", "density"), weight=("gas", "density"))

        w1 = ad.quantities.weighted_average_quantity(
            ("gas", "density"), ("gas", "density")
        )

        assert_equal(w, w1)

        w = ad.mean(("all", "particle_mass"))

        w1 = ad.quantities.weighted_average_quantity(
            ("all", "particle_mass"), ("all", "particle_ones")
        )

        assert_equal(w, w1)

        w = ad.mean(("all", "particle_mass"), weight=("all", "particle_mass"))

        w1 = ad.quantities.weighted_average_quantity(
            ("all", "particle_mass"), ("all", "particle_mass")
        )

        assert_equal(w, w1)

        # Projections
        p = ad.sum(("gas", "density"), axis=0)

        p1 = ds.proj(("gas", "density"), 0, data_source=ad, method="sum")

        assert_equal(p["gas", "density"], p1["gas", "density"])

        # Check by axis-name
        p = ad.sum(("gas", "density"), axis="x")

        assert_equal(p["gas", "density"], p1["gas", "density"])

        # Now we check proper projections
        p = ad.integrate(("gas", "density"), axis=0)
        p1 = ds.proj(("gas", "density"), 0, data_source=ad)

        assert_equal(p["gas", "density"], p1["gas", "density"])

        # Check by axis-name
        p = ad.integrate(("gas", "density"), axis="x")

        assert_equal(p["gas", "density"], p1["gas", "density"])


def test_min_max():
    for nprocs in [-1, 1, 2, 16]:
        fields = ["density", "temperature"]
        units = ["g/cm**3", "K"]
        if nprocs == -1:
            ds = fake_amr_ds(fields=fields, units=units, particles=20)
        else:
            ds = fake_random_ds(
                32, nprocs=nprocs, fields=fields, units=units, particles=20
            )

        ad = ds.all_data()

        q = ad.min(("gas", "density")).v
        assert_equal(q, ad["gas", "density"].min())

        q = ad.max(("gas", "density")).v
        assert_equal(q, ad["gas", "density"].max())

        q = ad.min(("all", "particle_mass")).v
        assert_equal(q, ad["all", "particle_mass"].min())

        q = ad.max(("all", "particle_mass")).v
        assert_equal(q, ad["all", "particle_mass"].max())

        ptp = ad.ptp(("gas", "density")).v
        assert_equal(ptp, ad["gas", "density"].max() - ad["gas", "density"].min())

        ptp = ad.ptp(("all", "particle_mass")).v
        assert_equal(
            ptp, ad["all", "particle_mass"].max() - ad["all", "particle_mass"].min()
        )

        p = ad.max(("gas", "density"), axis=1)
        p1 = ds.proj(("gas", "density"), 1, data_source=ad, method="max")
        assert_equal(p["gas", "density"], p1["gas", "density"])

        p = ad.min(("gas", "density"), axis=1)
        p1 = ds.proj(("gas", "density"), 1, data_source=ad, method="min")
        assert_equal(p["gas", "density"], p1["gas", "density"])

        p = ad.max(("gas", "density"), axis="y")
        p1 = ds.proj(("gas", "density"), 1, data_source=ad, method="max")
        assert_equal(p["gas", "density"], p1["gas", "density"])

        p = ad.min(("gas", "density"), axis="y")
        p1 = ds.proj(("gas", "density"), 1, data_source=ad, method="min")
        assert_equal(p["gas", "density"], p1["gas", "density"])

        # Test that we can get multiple in a single pass

        qrho, qtemp = ad.max([("gas", "density"), ("gas", "temperature")])
        assert_equal(qrho, ad["gas", "density"].max())
        assert_equal(qtemp, ad["gas", "temperature"].max())

        qrho, qtemp = ad.min([("gas", "density"), ("gas", "temperature")])
        assert_equal(qrho, ad["gas", "density"].min())
        assert_equal(qtemp, ad["gas", "temperature"].min())


def test_argmin():
    fields = ["density", "temperature"]
    units = ["g/cm**3", "K"]
    for nprocs in [-1, 1, 2, 16]:
        if nprocs == -1:
            ds = fake_amr_ds(fields=fields, units=units)
        else:
            ds = fake_random_ds(
                32,
                nprocs=nprocs,
                fields=fields,
                units=units,
            )

        ad = ds.all_data()

        q = ad.argmin(("gas", "density"), axis=[("gas", "density")])
        assert_equal(q, ad["gas", "density"].min())

        q1, q2 = ad.argmin(
            ("gas", "density"), axis=[("gas", "density"), ("gas", "temperature")]
        )
        mi = np.argmin(ad["gas", "density"])
        assert_equal(q1, ad["gas", "density"].min())
        assert_equal(q2, ad["gas", "temperature"][mi])

        pos = ad.argmin(("gas", "density"))
        mi = np.argmin(ad["gas", "density"])
        assert_equal(pos[0], ad["index", "x"][mi])
        assert_equal(pos[1], ad["index", "y"][mi])
        assert_equal(pos[2], ad["index", "z"][mi])


def test_argmax():
    fields = ["density", "temperature"]
    units = ["g/cm**3", "K"]
    for nprocs in [-1, 1, 2, 16]:
        if nprocs == -1:
            ds = fake_amr_ds(fields=fields, units=units)
        else:
            ds = fake_random_ds(
                32,
                nprocs=nprocs,
                fields=fields,
                units=units,
            )

        ad = ds.all_data()

        q = ad.argmax(("gas", "density"), axis=[("gas", "density")])
        assert_equal(q, ad["gas", "density"].max())

        q1, q2 = ad.argmax(
            ("gas", "density"), axis=[("gas", "density"), ("gas", "temperature")]
        )
        mi = np.argmax(ad["gas", "density"])
        assert_equal(q1, ad["gas", "density"].max())
        assert_equal(q2, ad["gas", "temperature"][mi])

        pos = ad.argmax(("gas", "density"))
        mi = np.argmax(ad["gas", "density"])
        assert_equal(pos[0], ad["index", "x"][mi])
        assert_equal(pos[1], ad["index", "y"][mi])
        assert_equal(pos[2], ad["index", "z"][mi])



"""
frontend API: a submodule that exposes user-facing defs and classes
"""

from .data_structures import AMRVACDataset, AMRVACGrid, AMRVACHierarchy
from .fields import AMRVACFieldInfo
from .io import AMRVACIOHandler, read_amrvac_namelist



import numpy as np

from yt.data_objects.index_subobjects.unstructured_mesh import UnstructuredMesh
from yt.data_objects.static_output import Dataset
from yt.data_objects.unions import MeshUnion
from yt.funcs import setdefaultattr
from yt.geometry.unstructured_mesh_handler import UnstructuredIndex
from yt.utilities.file_handler import NetCDF4FileHandler, valid_netcdf_signature
from yt.utilities.logger import ytLogger as mylog

from .fields import ExodusIIFieldInfo
from .util import get_num_pseudo_dims, load_info_records, sanitize_string


class ExodusIIUnstructuredMesh(UnstructuredMesh):
    _index_offset = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ExodusIIUnstructuredIndex(UnstructuredIndex):
    def __init__(self, ds, dataset_type="exodus_ii"):
        super().__init__(ds, dataset_type)

    def _initialize_mesh(self):
        coords = self.ds._read_coordinates()
        connectivity = self.ds._read_connectivity()
        self.meshes = []
        for mesh_id, conn_ind in enumerate(connectivity):
            displaced_coords = self.ds._apply_displacement(coords, mesh_id)
            mesh = ExodusIIUnstructuredMesh(
                mesh_id, self.index_filename, conn_ind, displaced_coords, self
            )
            self.meshes.append(mesh)
        self.mesh_union = MeshUnion("mesh_union", self.meshes)

    def _detect_output_fields(self):
        elem_names = self.dataset.parameters["elem_names"]
        node_names = self.dataset.parameters["nod_names"]
        fnames = elem_names + node_names
        self.field_list = []
        for i in range(1, len(self.meshes) + 1):
            self.field_list += [("connect%d" % i, fname) for fname in fnames]
        self.field_list += [("all", fname) for fname in fnames]


class ExodusIIDataset(Dataset):
    _load_requirements = ["netCDF4"]
    _index_class = ExodusIIUnstructuredIndex
    _field_info_class = ExodusIIFieldInfo

    def __init__(
        self,
        filename,
        step=0,
        displacements=None,
        dataset_type="exodus_ii",
        storage_filename=None,
        units_override=None,
    ):
        """

        A class used to represent an on-disk ExodusII dataset. The initializer takes
        two extra optional parameters, "step" and "displacements."

        Parameters
        ----------

        step : integer
            The step tells which time index to slice at. It throws an Error if
            the index is larger than the number of time outputs in the ExodusII
            file. Passing step=-1 picks out the last dataframe.
            Default is 0.

        displacements : dictionary of tuples
            This is a dictionary that controls whether or not displacement fields
            will be used with the meshes in this dataset. The keys of the
            displacements dictionary should the names of meshes in the file
            (e.g., "connect1", "connect2", etc... ), while the values should be
            tuples of the form (scale, offset), where "scale" is a floating point
            value and "offset" is an array-like with one component for each spatial
            dimension in the dataset. When the displacements for a given mesh are
            turned on, the coordinates of the vertices in that mesh get transformed
            as:

                  vertex_x = vertex_x + disp_x*scale + offset_x
                  vertex_y = vertex_y + disp_y*scale + offset_y
                  vertex_z = vertex_z + disp_z*scale + offset_z

            If no displacement
            fields (assumed to be named 'disp_x', 'disp_y', etc... ) are detected in
            the output file, then this dictionary is ignored.

        Examples
        --------

        This will load the Dataset at time index '0' with displacements turned off.

        >>> import yt
        >>> ds = yt.load("MOOSE_sample_data/mps_out.e")

        This will load the Dataset at the final index with displacements turned off.

        >>> import yt
        >>> ds = yt.load("MOOSE_sample_data/mps_out.e", step=-1)

        This will load the Dataset at index 10, turning on displacement fields for
        the 2nd mesh without applying any scale or offset:

        >>> import yt
        >>> ds = yt.load(
        ...     "MOOSE_sample_data/mps_out.e",
        ...     step=10,
        ...     displacements={"connect2": (1.0, [0.0, 0.0, 0.0])},
        ... )

        This will load the Dataset at index 10, scaling the displacements
        in the 2nd mesh by a factor of 5 while not applying an offset:

        >>> import yt
        >>> ds = yt.load(
        ...     "MOOSE_sample_data/mps_out.e",
        ...     step=10,
        ...     displacements={"connect2": (5.0, [0.0, 0.0, 0.0])},
        ... )

        This will load the Dataset at index 10, scaling the displacements for
        the 2nd mesh by a factor of 5.0 and shifting all the vertices in
        the first mesh by 1.0 unit in the z direction.

        >>> import yt
        >>> ds = yt.load(
        ...     "MOOSE_sample_data/mps_out.e",
        ...     step=10,
        ...     displacements={
        ...         "connect1": (0.0, [0.0, 0.0, 1.0]),
        ...         "connect2": (5.0, [0.0, 0.0, 0.0]),
        ...     },
        ... )

        """
        self.step = step
        if displacements is None:
            self.displacements = {}
        else:
            self.displacements = displacements
        self.storage_filename = storage_filename

        super().__init__(filename, dataset_type, units_override=units_override)

        self.fluid_types += self._get_fluid_types()
        self.default_field = [f for f in self.field_list if f[0] == "connect1"][-1]

    @property
    def index_filename(self):
        # historic alias
        return self.filename

    def _set_code_unit_attributes(self):
        # This is where quantities are created that represent the various
        # on-disk units.  These are the currently available quantities which
        # should be set, along with examples of how to set them to standard
        # values.
        #
        setdefaultattr(self, "length_unit", self.quan(1.0, "cm"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "g"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))
        #
        # These can also be set:
        # self.velocity_unit = self.quan(1.0, "cm/s")
        # self.magnetic_unit = self.quan(1.0, "gauss")

    def _parse_parameter_file(self):
        self._handle = NetCDF4FileHandler(self.parameter_filename)
        with self._handle.open_ds() as ds:
            self._read_glo_var()
            self.dimensionality = ds.variables["coor_names"].shape[0]
            self.parameters["info_records"] = self._load_info_records()
            self.num_steps = len(ds.variables["time_whole"])
            self.current_time = self._get_current_time()
            self.parameters["num_meshes"] = ds.variables["eb_status"].shape[0]
            self.parameters["elem_names"] = self._get_elem_names()
            self.parameters["nod_names"] = self._get_nod_names()
            self.domain_left_edge, self.domain_right_edge = self._load_domain_edge()
            self._periodicity = (False, False, False)

        # These attributes don't really make sense for unstructured
        # mesh data, but yt warns if they are not present, so we set
        # them to dummy values here.
        self.domain_dimensions = np.ones(3, "int32")
        self.cosmological_simulation = 0
        self.current_redshift = 0
        self.omega_lambda = 0
        self.omega_matter = 0
        self.hubble_constant = 0
        self.refine_by = 0

    def _get_fluid_types(self):
        with NetCDF4FileHandler(self.parameter_filename).open_ds() as ds:
            fluid_types = ()
            i = 1
            while True:
                ftype = "connect%d" % i
                if ftype in ds.variables:
                    fluid_types += (ftype,)
                    i += 1
                else:
                    break
            fluid_types += ("all",)
            return fluid_types

    def _read_glo_var(self):
        """
        Adds each global variable to the dict of parameters

        """
        names = self._get_glo_names()
        if not names:
            return
        with self._handle.open_ds() as ds:
            values = ds.variables["vals_glo_var"][:].transpose()
            for name, value in zip(names, values, strict=True):
                self.parameters[name] = value

    def _load_info_records(self):
        """
        Returns parsed version of the info_records.
        """
        with self._handle.open_ds() as ds:
            try:
                return load_info_records(ds.variables["info_records"])
            except (KeyError, TypeError):
                mylog.warning("No info_records found")
                return []

    def _get_current_time(self):
        with self._handle.open_ds() as ds:
            try:
                return ds.variables["time_whole"][self.step]
            except IndexError as e:
                raise RuntimeError(
                    "Invalid step number, max is %d" % (self.num_steps - 1)
                ) from e
            except (KeyError, TypeError):
                return 0.0

    def _get_glo_names(self):
        """

        Returns the names of the global vars, if available.

        """

        with self._handle.open_ds() as ds:
            if "name_glo_var" not in ds.variables:
                mylog.warning("name_glo_var not found")
                return []
            else:
                return [
                    sanitize_string(v.tobytes()) for v in ds.variables["name_glo_var"]
                ]

    def _get_elem_names(self):
        """

        Returns the names of the element vars, if available.

        """

        with self._handle.open_ds() as ds:
            if "name_elem_var" not in ds.variables:
                mylog.warning("name_elem_var not found")
                return []
            else:
                return [
                    sanitize_string(v.tobytes()) for v in ds.variables["name_elem_var"]
                ]

    def _get_nod_names(self):
        """

        Returns the names of the node vars, if available

        """

        with self._handle.open_ds() as ds:
            if "name_nod_var" not in ds.variables:
                mylog.warning("name_nod_var not found")
                return []
            else:
                return [
                    sanitize_string(v.tobytes()) for v in ds.variables["name_nod_var"]
                ]

    def _read_coordinates(self):
        """

        Loads the coordinates for the mesh

        """

        coord_axes = "xyz"[: self.dimensionality]

        mylog.info("Loading coordinates")
        with self._handle.open_ds() as ds:
            if "coord" not in ds.variables:
                coords = (
                    np.array([ds.variables[f"coord{ax}"][:] for ax in coord_axes])
                    .transpose()
                    .astype("f8")
                )
            else:
                coords = (
                    np.array(list(ds.variables["coord"][:])).transpose().astype("f8")
                )
            return coords

    def _apply_displacement(self, coords, mesh_id):
        mesh_name = "connect%d" % (mesh_id + 1)
        new_coords = coords.copy()
        if mesh_name not in self.displacements:
            return new_coords
        fac = self.displacements[mesh_name][0]
        offset = self.displacements[mesh_name][1]

        coord_axes = "xyz"[: self.dimensionality]
        with self._handle.open_ds() as ds:
            for i, ax in enumerate(coord_axes):
                if f"disp_{ax}" in self.parameters["nod_names"]:
                    ind = self.parameters["nod_names"].index(f"disp_{ax}")
                    disp = ds.variables["vals_nod_var%d" % (ind + 1)][self.step]
                    new_coords[:, i] = coords[:, i] + fac * disp + offset[i]

            return new_coords

    def _read_connectivity(self):
        """
        Loads the connectivity data for the mesh
        """
        mylog.info("Loading connectivity")
        connectivity = []
        with self._handle.open_ds() as ds:
            for i in range(self.parameters["num_meshes"]):
                var = ds.variables["connect%d" % (i + 1)][:].astype("i8")
                try:
                    elem_type = var.elem_type.lower()
                    if elem_type == "nfaced":
                        raise NotImplementedError(
                            "3D arbitrary polyhedra are not implemented yet"
                        )
                    arbitrary_polyhedron = elem_type == "nsided"
                except AttributeError:
                    arbitrary_polyhedron = False

                conn = var[:]
                if arbitrary_polyhedron:
                    nodes_per_element = ds.variables[f"ebepecnt{i + 1}"]
                    npe = nodes_per_element[0]
                    if np.any(nodes_per_element != npe):
                        raise NotImplementedError("only equal-size polyhedra supported")
                    q, r = np.divmod(len(conn), npe)
                    assert r == 0
                    conn = conn.reshape(q, npe)
                connectivity.append(conn)
            return connectivity

    def _load_domain_edge(self):
        """
        Loads the boundaries for the domain edge

        """

        coords = self._read_coordinates()
        connectivity = self._read_connectivity()

        mi = 1e300
        ma = -1e300
        for mesh_id, _ in enumerate(connectivity):
            displaced_coords = self._apply_displacement(coords, mesh_id)
            mi = np.minimum(displaced_coords.min(axis=0), mi)
            ma = np.maximum(displaced_coords.max(axis=0), ma)

        # pad domain boundaries
        width = ma - mi
        mi -= 0.1 * width
        ma += 0.1 * width

        # set up pseudo-3D for lodim datasets here
        for _ in range(self.dimensionality, 3):
            mi = np.append(mi, 0.0)
            ma = np.append(ma, 1.0)

        num_pseudo_dims = get_num_pseudo_dims(coords)
        self.dimensionality -= num_pseudo_dims
        for i in range(self.dimensionality, 3):
            mi[i] = 0.0
            ma[i] = 1.0

        return mi, ma

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if not valid_netcdf_signature(filename):
            return False

        if cls._missing_load_requirements():
            return False

        try:
            from netCDF4 import Dataset

            # We use keepweakref here to avoid holding onto the file handle
            # which can interfere with other is_valid calls.
            with Dataset(filename, keepweakref=True) as f:
                f.variables["connect1"]
            return True
        except Exception:
            return False



from collections import defaultdict
from itertools import product

import pytest

from yt.testing import fake_random_ds
from yt.utilities.exceptions import YTFieldNotFound


def test_unexisting_field_access():
    ds = fake_random_ds(16, particles=10)

    fname2ftype = defaultdict(list)

    for ft, fn in ds.derived_field_list:
        fname2ftype[fn].append(ft)

    ad = ds.all_data()

    ftypes = ("gas", "io")
    fnames = (
        "density",
        "particle_position_x",
        "particle_position_y",
        "particle_position_z",
    )

    # Try invalid ftypes, fnames combinations
    for ft, fn in product(ftypes, fnames):
        if (ft, fn) in ds.derived_field_list:
            continue

        with pytest.raises(YTFieldNotFound) as excinfo:
            ad[ft, fn]

        # Make sure the existing field has been suggested
        for possible_ft in fname2ftype[fn]:
            assert str((possible_ft, fn)) in str(excinfo.value)

    # Try typos
    for bad_field, good_field in (
        (("gas", "densi_y"), ("gas", "density")),
        (("oi", "particle_mass"), ("io", "particle_mass")),
        (("gas", "DENSITY"), ("gas", "density")),
    ):
        with pytest.raises(YTFieldNotFound) as excinfo:
            ad[bad_field]

        assert str(good_field) in str(excinfo.value)



from . import tests
from .data_structures import (
    EnzoDataset,
    EnzoDatasetInMemory,
    EnzoGrid,
    EnzoGridInMemory,
    EnzoHierarchy,
    EnzoHierarchy1D,
    EnzoHierarchy2D,
    EnzoHierarchyInMemory,
)
from .fields import EnzoFieldInfo
from .io import (
    IOHandlerInMemory,
    IOHandlerPacked1D,
    IOHandlerPacked2D,
    IOHandlerPackedHDF5,
)
from .simulation_handling import EnzoSimulation

add_enzo_field = EnzoFieldInfo.add_field



"""
CF-Radial frontend tests



"""

import os
import shutil
import tempfile

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from yt.frontends.cf_radial.data_structures import CFRadialDataset
from yt.testing import (
    requires_file,
    requires_module,
    units_override_check,
)
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

cf = "CfRadialGrid/grid1.nc"  # an already gridded cfradial file
cf_nongridded = (
    "CfRadialGrid/swx_20120520_0641.nc"  # cfradial file without cartesian grid
)
_fields_cfradial = ["reflectivity", "velocity", "gate_id", "differential_phase", "ROI"]

_fields_units = {
    "reflectivity": "dimensionless",
    "velocity": "m/s",
    "differential_phase": "degree",
    "gate_id": "dimensionless",
    "ROI": "m",
}


@requires_module("xarray")
@requires_file(cf)
def test_units_override():
    units_override_check(cf)


@requires_module("xarray")
@requires_file(cf)
def test_cf_radial_gridded():
    ds = data_dir_load(cf)
    assert isinstance(ds, CFRadialDataset)

    check_domain(ds)
    check_origin_latitude_longitude(ds)
    ad = ds.all_data()

    for field in _fields_cfradial:
        check_fields(ds, field)
        check_field_units(ad, field, _fields_units[field])


def check_fields(ds, field):
    assert ("cf_radial", field) in ds.field_list
    with ds._handle() as xr_ds_handle:
        assert field in xr_ds_handle.variables.keys()


def check_field_units(ad, field, value):
    assert str(ad["cf_radial", field].units) == value


def check_origin_latitude_longitude(ds):
    assert_almost_equal(ds.origin_latitude.values, 36.49120001)
    assert_almost_equal(ds.origin_longitude.values, -97.5939)


def check_domain(ds):
    domain_dim_array = [251, 251, 46]
    assert_equal(ds.domain_dimensions, domain_dim_array)

    domain_center_array = [0.0, 0.0, 7500.0]
    assert_equal(ds.domain_center, domain_center_array)

    domain_left_array = [-50000.0, -50000.0, 0.0]
    assert_equal(ds.domain_left_edge, domain_left_array)

    domain_right_array = [50000.0, 50000.0, 15000.0]
    assert_equal(ds.domain_right_edge, domain_right_array)


@requires_module("xarray")
@requires_file(cf_nongridded)
def test_auto_gridding():
    # loads up a radial dataset, which triggers the gridding.

    # create temporary directory and grid file
    tempdir = tempfile.mkdtemp()
    grid_file = os.path.join(tempdir, "temp_grid.nc")

    # this load will trigger the re-gridding and write out the gridded file
    # from which data will be loaded. With default gridding params, this takes
    # on the order of 10s, but since we are not testing actual output here, we
    # can decrease the resolution to speed it up.
    grid_shape = (10, 10, 10)
    ds = data_dir_load(
        cf_nongridded, kwargs={"storage_filename": grid_file, "grid_shape": grid_shape}
    )
    assert os.path.exists(grid_file)

    # check that the cartesian fields exist now
    with ds._handle() as xr_ds_handle:
        on_disk_fields = xr_ds_handle.variables.keys()
        for field in ["x", "y", "z"]:
            assert field in on_disk_fields

    assert all(ds.domain_dimensions == grid_shape)

    # check that we can load the gridded file too
    ds = data_dir_load(grid_file)
    assert isinstance(ds, CFRadialDataset)

    shutil.rmtree(tempdir)


@requires_module("xarray")
@requires_file(cf_nongridded)
def test_grid_parameters():
    # checks that the gridding parameters are used and that conflicts in parameters
    # are resolved as expected.
    tempdir = tempfile.mkdtemp()
    grid_file = os.path.join(tempdir, "temp_grid_params.nc")

    # check that the grid parameters work
    cfkwargs = {
        "storage_filename": grid_file,
        "grid_shape": (10, 10, 10),
        "grid_limit_x": (-10000, 10000),
        "grid_limit_y": (-10000, 10000),
        "grid_limit_z": (500, 20000),
    }
    ds = data_dir_load(cf_nongridded, kwargs=cfkwargs)

    expected_width = []
    for dim in "xyz":
        minval, maxval = cfkwargs[f"grid_limit_{dim}"]
        expected_width.append(maxval - minval)
    expected_width = np.array(expected_width)

    actual_width = ds.domain_width.to_value("m")
    assert all(expected_width == actual_width)
    assert all(ds.domain_dimensions == cfkwargs["grid_shape"])

    # check the grid parameter conflicts

    # on re-load with default grid params it will reload storage_filename if
    # it exists. Just checking that this runs...
    _ = data_dir_load(cf_nongridded, kwargs={"storage_filename": grid_file})

    # if storage_filename exists, grid parameters are ignored (with a warning)
    # and the domain_dimensions will match the original
    new_kwargs = {"storage_filename": grid_file, "grid_shape": (15, 15, 15)}
    ds = data_dir_load(cf_nongridded, kwargs=new_kwargs)
    assert all(ds.domain_dimensions == cfkwargs["grid_shape"])

    # if we overwrite, the regridding should run and the dimensions should match
    # the desired dimensions
    new_kwargs["storage_overwrite"] = True
    ds = data_dir_load(cf_nongridded, kwargs=new_kwargs)
    assert all(ds.domain_dimensions == new_kwargs["grid_shape"])

    shutil.rmtree(tempdir)


@requires_module("xarray")
@requires_ds(cf)
def test_cfradial_grid_field_values():
    ds = data_dir_load(cf)
    fields_to_check = [("cf_radial", field) for field in _fields_cfradial]
    wtfield = ("cf_radial", "reflectivity")
    for test in small_patch_amr(
        ds, fields_to_check, input_center=ds.domain_center, input_weight=wtfield
    ):
        test_cfradial_grid_field_values.__name__ = test.description
        yield test



import numpy as np

from yt.utilities.operator_registry import OperatorRegistry

clump_info_registry = OperatorRegistry()


def add_clump_info(name, function):
    clump_info_registry[name] = ClumpInfoCallback(name, function)


class ClumpInfoCallback:
    r"""
    A ClumpInfoCallback is a function that takes a clump, computes a
    quantity, and returns a string to be printed out for writing clump info.
    """

    def __init__(self, name, function, args=None, kwargs=None):
        self.name = name
        self.function = function
        self.args = args
        if self.args is None:
            self.args = []
        self.kwargs = kwargs
        if self.kwargs is None:
            self.kwargs = {}

    def __call__(self, clump):
        if self.name not in clump.info:
            clump.info[self.name] = self.function(clump, *self.args, **self.kwargs)
        rv = clump.info[self.name]
        return rv[0] % rv[1]


def _center_of_mass(clump, units="code_length", **kwargs):
    p = clump.quantities.center_of_mass(**kwargs)
    return "Center of mass: %s.", p.to(units)


add_clump_info("center_of_mass", _center_of_mass)


def _total_cells(clump):
    n_cells = clump.data["index", "ones"].size
    return "Cells: %d.", n_cells


add_clump_info("total_cells", _total_cells)


def _cell_mass(clump):
    cell_mass = clump.data["gas", "cell_mass"].sum().in_units("Msun")
    return "Mass: %e Msun.", cell_mass


add_clump_info("cell_mass", _cell_mass)


def _mass_weighted_jeans_mass(clump):
    jeans_mass = clump.data.quantities.weighted_average_quantity(
        "jeans_mass", ("gas", "cell_mass")
    ).in_units("Msun")
    return "Jeans Mass (mass-weighted): %.6e Msolar.", jeans_mass


add_clump_info("mass_weighted_jeans_mass", _mass_weighted_jeans_mass)


def _volume_weighted_jeans_mass(clump):
    jeans_mass = clump.data.quantities.weighted_average_quantity(
        "jeans_mass", ("index", "cell_volume")
    ).in_units("Msun")
    return "Jeans Mass (volume-weighted): %.6e Msolar.", jeans_mass


add_clump_info("volume_weighted_jeans_mass", _volume_weighted_jeans_mass)


def _max_grid_level(clump):
    max_level = clump.data["index", "grid_level"].max()
    return "Max grid level: %d.", max_level


add_clump_info("max_grid_level", _max_grid_level)


def _min_number_density(clump):
    min_n = clump.data["gas", "number_density"].min().in_units("cm**-3")
    return "Min number density: %.6e cm^-3.", min_n


add_clump_info("min_number_density", _min_number_density)


def _max_number_density(clump):
    max_n = clump.data["gas", "number_density"].max().in_units("cm**-3")
    return "Max number density: %.6e cm^-3.", max_n


add_clump_info("max_number_density", _max_number_density)


def _distance_to_main_clump(clump, units="pc"):
    master = clump
    while master.parent is not None:
        master = master.parent
    master_com = clump.data.ds.arr(master.data.quantities.center_of_mass())
    my_com = clump.data.ds.arr(clump.data.quantities.center_of_mass())
    distance = np.sqrt(((master_com - my_com) ** 2).sum())
    distance.convert_to_units("pc")
    return (
        f"Distance from master center of mass: %.6e {units}.",
        distance.in_units(units),
    )


add_clump_info("distance_to_main_clump", _distance_to_main_clump)
